# pylint: disable=all
import collections
import dataclasses
import json
import logging
import os
import subprocess
import time
from collections.abc import Generator
from datetime import datetime, timezone
from pathlib import Path
from typing import TextIO, List

import sqlglot
from sqlglot.expressions import Expression
from sqlglot.dialects.dialect import Dialect
from sqlglot.dialects.databricks import Databricks
from sqlglot.errors import ErrorLevel

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ReportEntry:
    project: str
    commit_hash: str | None
    version: str
    timestamp: str
    source_dialect: str
    target_dialect: str
    file: str
    parsed: int = 0  # 1 for success, 0 for failure
    statements: int = 0  # number of statements parsed
    transpiled: int = 0  # 1 for success, 0 for failure
    transpiled_statements: int = 0  # number of statements transpiled
    failures: List[dict] = dataclasses.field(default_factory=lambda: [])


def sqlglot_run_coverage(dialect, subfolder):
    input_dir = get_env_var("INPUT_DIR_PARENT", required=True)
    output_dir = get_env_var("OUTPUT_DIR", required=True)
    sqlglot_version = sqlglot.__version__
    SQLGLOT_COMMIT_HASH = ""  # C0103 pylint

    if not input_dir:
        raise ValueError("Environment variable `INPUT_DIR_PARENT` is required")
    if not output_dir:
        raise ValueError("Environment variable `OUTPUT_DIR` is required")

    collect_transpilation_stats(
        "SQLGlot",
        SQLGLOT_COMMIT_HASH,
        sqlglot_version,
        dialect,
        Databricks,
        Path(input_dir) / subfolder,
        Path(output_dir),
    )


def local_report(output_dir: Path):
    all = collections.defaultdict(list)
    for file in output_dir.rglob("*.json"):
        with file.open("r", encoding="utf8") as f:
            for line in f:
                raw = json.loads(line)
                entry = ReportEntry(**raw)
                all[(entry.project, entry.source_dialect)].append(entry)
    for (project, dialect), entries in sorted(all.items()):
        total = len(entries)
        parsed = sum(entry.parsed for entry in entries)
        transpiled = sum(entry.transpiled for entry in entries)
        parse_ratio = parsed / total
        transpile_ratio = transpiled / total
        print(
            f"{project} -> {dialect}: {parse_ratio:.2%} parsed ({parsed}/{total}), "
            f"{transpile_ratio:.2%} transpiled ({transpiled}/{total})"
        )


def get_supported_sql_files(input_dir: Path) -> Generator[Path, None, None]:
    yield from filter(lambda item: item.is_file() and item.suffix.lower() in [".sql", ".ddl"], input_dir.rglob("*"))


def write_json_line(file: TextIO, content: ReportEntry):
    json.dump(dataclasses.asdict(content), file)
    file.write("\n")


def get_env_var(env_var: str, *, required: bool = False) -> str | None:
    """
    Get the value of an environment variable.

    :param env_var: The name of the environment variable to get the value of.
    :param required: Indicates if the environment variable is required and raises a ValueError if it's not set.
    :return: Returns the environment variable's value, or None if it's not set and not required.
    """
    value = os.getenv(env_var)
    if value is None and required:
        message = f"Environment variable {env_var} is not set"
        raise ValueError(message)
    return value


def get_current_commit_hash() -> str | None:
    try:
        return (
            subprocess.check_output(
                ["/usr/bin/git", "rev-parse", "--short", "HEAD"],
                cwd=Path(__file__).resolve().parent,
            )
            .decode("ascii")
            .strip()
        )
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        logger.warning(f"Could not get the current commit hash. {e!s}")
        return None


def get_current_time_utc() -> datetime:
    return datetime.now(timezone.utc)


def parse_sql(sql: str, dialect: type[Dialect]) -> list[Expression]:
    return [
        expression for expression in sqlglot.parse(sql, read=dialect, error_level=ErrorLevel.IMMEDIATE) if expression
    ]


def generate_sql(expressions: list[Expression], dialect: type[Dialect]) -> list[str]:
    generator_dialect = Dialect.get_or_raise(dialect)
    return [generator_dialect.generate(expression, copy=False) for expression in expressions if expression]


def _ensure_valid_io_paths(input_dir: Path, result_dir: Path):
    if not input_dir.exists() or not input_dir.is_dir():
        message = f"The input path {input_dir} doesn't exist or is not a directory"
        raise NotADirectoryError(message)

    if not result_dir.exists():
        logger.info(f"Creating the output directory {result_dir}")
        result_dir.mkdir(parents=True)
    elif not result_dir.is_dir():
        message = f"The output path {result_dir} exists but is not a directory"
        raise NotADirectoryError(message)


def _get_report_file_path(
    project: str,
    source_dialect: type[Dialect],
    target_dialect: type[Dialect],
    result_dir: Path,
) -> Path:
    source_dialect_name = source_dialect.__name__
    target_dialect_name = target_dialect.__name__
    current_time_ns = time.time_ns()
    return result_dir / f"{project}_{source_dialect_name}_{target_dialect_name}_{current_time_ns}.json".lower()


def _prepare_report_entry(
    project: str,
    commit_hash: str,
    version: str,
    source_dialect: type[Dialect],
    target_dialect: type[Dialect],
    file_path: str,
    sql: str,
) -> ReportEntry:
    report_entry = ReportEntry(
        project=project,
        commit_hash=commit_hash,
        version=version,
        timestamp=get_current_time_utc().isoformat(),
        source_dialect=source_dialect.__name__,
        target_dialect=target_dialect.__name__,
        file=file_path,
    )
    try:
        expressions = parse_sql(sql, source_dialect)
        report_entry.parsed = 1
        report_entry.statements = len(expressions)
    except Exception as pe:
        report_entry.failures.append({'error_code': type(pe).__name__, 'error_message': repr(pe)})
        return report_entry

    try:
        generated_sqls = generate_sql(expressions, target_dialect)
        report_entry.transpiled = 1
        report_entry.transpiled_statements = len([sql for sql in generated_sqls if sql.strip()])
    except Exception as te:
        report_entry.failures.append({'error_code': type(te).__name__, 'error_message': repr(te)})

    return report_entry


def collect_transpilation_stats(
    project: str,
    commit_hash: str,
    version: str,
    source_dialect: type[Dialect],
    target_dialect: type[Dialect],
    input_dir: Path,
    result_dir: Path,
):
    _ensure_valid_io_paths(input_dir, result_dir)
    report_file_path = _get_report_file_path(project, source_dialect, target_dialect, result_dir)

    with report_file_path.open("w", encoding="utf8") as report_file:
        for input_file in get_supported_sql_files(input_dir):
            with input_file.open("r", encoding="utf-8-sig") as file:
                sql = file.read()

            file_path = str(input_file.absolute().relative_to(input_dir.parent.absolute()))
            report_entry = _prepare_report_entry(
                project,
                commit_hash,
                version,
                source_dialect,
                target_dialect,
                file_path,
                sql,
            )
            write_json_line(report_file, report_entry)



from databricks.labs.blueprint.logger import install_logger
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.sdk import WorkspaceClient
from databricks.sdk.core import with_user_agent_extra

from databricks.labs.lakebridge import __version__
from databricks.labs.lakebridge.install import installer as _installer
from databricks.labs.lakebridge.transpiler.repository import TranspilerRepository


def main() -> None:
    install_logger()
    with_user_agent_extra("cmd", "install")

    logger = get_logger(__file__)
    logger.setLevel("INFO")

    installer = _installer(
        WorkspaceClient(product="lakebridge", product_version=__version__),
        transpiler_repository=TranspilerRepository.user_home(),
    )
    if not installer.upgrade_installed_transpilers():
        logger.debug("No existing Lakebridge transpilers detected; assuming fresh installation.")

    logger.info("Successfully Setup Lakebridge Components Locally")
    logger.info("For more information, please visit https://databrickslabs.github.io/lakebridge/")


if __name__ == "__main__":
    main()



from collections.abc import Callable
from functools import partial

from pyspark.sql.types import DataType, NumericType
from sqlglot import Dialect
from sqlglot import expressions as exp

from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_dialect
from databricks.labs.lakebridge.reconcile.recon_config import HashAlgoMapping


def _apply_func_expr(expr: exp.Expression, expr_func: Callable, **kwargs) -> exp.Expression:
    is_terminal = isinstance(expr, exp.Column)
    new_expr = expr.copy()
    for node in new_expr.dfs():
        if isinstance(node, exp.Column):
            column_name = node.name
            table_name = node.table
            func = expr_func(this=exp.Column(this=column_name, table=table_name), **kwargs)
            if is_terminal:
                return func
            node.replace(func)
    return new_expr


def concat(expr: list[exp.Expression]) -> exp.Expression:
    return exp.Concat(expressions=expr, safe=True)


def sha2(expr: exp.Expression, num_bits: str, is_expr: bool = False) -> exp.Expression:
    if is_expr:
        return exp.SHA2(this=expr, length=exp.Literal(this=num_bits, is_string=False))
    return _apply_func_expr(expr, exp.SHA2, length=exp.Literal(this=num_bits, is_string=False))


def md5(expr: exp.Expression, is_expr: bool = False) -> exp.Expression:
    if is_expr:
        return exp.MD5(this=expr)
    return _apply_func_expr(expr, exp.MD5)


def lower(expr: exp.Expression, is_expr: bool = False) -> exp.Expression:
    if is_expr:
        return exp.Lower(this=expr)
    return _apply_func_expr(expr, exp.Lower)


def coalesce(expr: exp.Expression, default="0", is_string=False) -> exp.Expression:
    expressions = [exp.Literal(this=default, is_string=is_string)]
    return _apply_func_expr(expr, exp.Coalesce, expressions=expressions)


def trim(expr: exp.Expression) -> exp.Trim | exp.Expression:
    return _apply_func_expr(expr, exp.Trim)


def json_format(expr: exp.Expression, options: dict[str, str] | None = None) -> exp.Expression:
    return _apply_func_expr(expr, exp.JSONFormat, options=options)


def sort_array(expr: exp.Expression, asc=True) -> exp.Expression:
    return _apply_func_expr(expr, exp.SortArray, asc=exp.Boolean(this=asc))


def to_char(expr: exp.Expression, to_format=None, nls_param=None) -> exp.Expression:
    if to_format:
        return _apply_func_expr(
            expr, exp.ToChar, format=exp.Literal(this=to_format, is_string=True), nls_param=nls_param
        )
    return _apply_func_expr(expr, exp.ToChar)


def array_to_string(
    expr: exp.Expression,
    delimiter: str = ",",
    is_string=True,
    null_replacement: str | None = None,
    is_null_replace=True,
) -> exp.Expression:
    if null_replacement:
        return _apply_func_expr(
            expr,
            exp.ArrayToString,
            expression=[exp.Literal(this=delimiter, is_string=is_string)],
            null=exp.Literal(this=null_replacement, is_string=is_null_replace),
        )
    return _apply_func_expr(expr, exp.ArrayToString, expression=[exp.Literal(this=delimiter, is_string=is_string)])


def array_sort(expr: exp.Expression, asc=True) -> exp.Expression:
    return _apply_func_expr(expr, exp.ArraySort, expression=exp.Boolean(this=asc))


def anonymous(expr: exp.Column, func: str, is_expr: bool = False, dialect=None) -> exp.Expression:
    """

    This function used in cases where the sql functions are not available in sqlGlot expressions
    Example:
        >>> from sqlglot import parse_one
        >>> print(repr(parse_one('select unix_timestamp(col1)')))

    the above code gives you a Select Expression of Anonymous function.

    To achieve the same,we can use the function as below:
    eg:
        >>> expr = parse_one("select col1 from dual")
        >>> transformed_expr=anonymous(expr,"unix_timestamp({})")
        >>> print(transformed_expr)
        'SELECT UNIX_TIMESTAMP(col1) FROM DUAL'

    """
    if is_expr:
        if dialect:
            return exp.Column(this=func.format(expr.sql(dialect=dialect)))
        return exp.Column(this=func.format(expr))
    is_terminal = isinstance(expr, exp.Column)
    new_expr = expr.copy()
    for node in new_expr.dfs():
        if isinstance(node, exp.Column):
            name = f"{node.table}.{node.name}" if node.table else node.name
            anonymous_func = exp.Column(this=func.format(name))
            if is_terminal:
                return anonymous_func
            node.replace(anonymous_func)
    return new_expr


# TODO Standardize impl and use quoted and Identifier/Column consistently
def build_column(this: exp.ExpOrStr, table_name="", quoted=False, alias=None) -> exp.Expression:
    if alias:
        if isinstance(this, str):
            return exp.Alias(
                this=exp.Column(this=this, table=table_name), alias=exp.Identifier(this=alias, quoted=quoted)
            )
        return exp.Alias(this=this, alias=exp.Identifier(this=alias, quoted=quoted))
    return exp.Column(this=exp.Identifier(this=this, quoted=quoted), table=table_name)


def build_column_no_alias(this: str, table_name="") -> exp.Expression:
    return exp.Column(this=this, table=table_name)


def build_literal(this: exp.ExpOrStr, alias=None, quoted=False, is_string=True, cast=None) -> exp.Expression:
    base_literal = exp.Literal(this=this, is_string=is_string)
    if not cast and not alias:
        return base_literal

    cast_expr = exp.Cast(this=base_literal, to=exp.DataType(this=cast)) if cast else base_literal
    return exp.Alias(this=cast_expr, alias=exp.Identifier(this=alias, quoted=quoted)) if alias else cast_expr


def transform_expression(
    expr: exp.Expression,
    funcs: list[Callable[[exp.Expression], exp.Expression]],
) -> exp.Expression:
    for func in funcs:
        expr = func(expr)
    assert isinstance(expr, exp.Expression), (
        f"Func returned an instance of type [{type(expr)}], " "should have been Expression."
    )
    return expr


def get_hash_transform(
    source: Dialect,
    layer: str,
):
    dialect_algo = Dialect_hash_algo_mapping.get(source)
    if not dialect_algo:
        raise ValueError(f"Source {source} is not supported. Please add it to Dialect_hash_algo_mapping dictionary.")

    layer_algo = getattr(dialect_algo, layer, None)
    if not layer_algo:
        raise ValueError(
            f"Layer {layer} is not supported for source {source}. Please add it to Dialect_hash_algo_mapping dictionary."
        )
    return [layer_algo]


def build_from_clause(table_name: str, table_alias: str | None = None) -> exp.From:
    return exp.From(this=exp.Table(this=exp.Identifier(this=table_name), alias=table_alias))


def build_join_clause(
    table_name: str,
    join_columns: list,
    source_table_alias: str | None = None,
    target_table_alias: str | None = None,
    kind: str = "inner",
    func: Callable = exp.NullSafeEQ,
) -> exp.Join:
    join_conditions = []
    for column in join_columns:
        join_condition = func(
            this=exp.Column(this=column, table=source_table_alias),
            expression=exp.Column(this=column, table=target_table_alias),
        )
        join_conditions.append(join_condition)

    # Combine all join conditions with AND
    on_condition: exp.NullSafeEQ | exp.And = join_conditions[0]
    for condition in join_conditions[1:]:
        on_condition = exp.And(this=on_condition, expression=condition)

    return exp.Join(
        this=exp.Table(this=exp.Identifier(this=table_name), alias=target_table_alias), kind=kind, on=on_condition
    )


def build_sub(
    left_column_name: str,
    right_column_name: str,
    left_table_name: str | None = None,
    right_table_name: str | None = None,
    quoted: bool = False,
) -> exp.Sub:
    return exp.Sub(
        this=build_column(left_column_name, left_table_name, quoted=quoted),
        expression=build_column(right_column_name, right_table_name, quoted=quoted),
    )


def build_where_clause(where_clause: list[exp.Expression], condition_type: str = "or") -> exp.Expression:
    func = exp.Or if condition_type == "or" else exp.And
    # Start with a default
    combined_expression: exp.Expression = exp.Paren(this=func(this='1 = 1', expression='1 = 1'))

    # Loop through the expressions and combine them with OR
    for expression in where_clause:
        combined_expression = func(this=combined_expression, expression=expression)

    return combined_expression


def build_if(this: exp.Expression, true: exp.Expression, false: exp.Expression | None = None) -> exp.If:
    return exp.If(this=this, true=true, false=false)


def build_between(this: exp.Expression, low: exp.Expression, high: exp.Expression) -> exp.Between:
    return exp.Between(this=this, low=low, high=high)


def _get_is_string(column_types_dict: dict[str, DataType], column_name: str) -> bool:
    if isinstance(column_types_dict.get(column_name), NumericType):
        return False
    return True


DataType_transform_mapping: dict[str, dict[str, list[partial[exp.Expression]]]] = {
    "universal": {"default": [partial(coalesce, default='_null_recon_', is_string=True), partial(trim)]},
    "snowflake": {exp.DataType.Type.ARRAY.value: [partial(array_to_string), partial(array_sort)]},
    "oracle": {
        exp.DataType.Type.NCHAR.value: [
            partial(anonymous, func="NVL(TRIM(TO_CHAR({})),'_null_recon_')", dialect=get_dialect("oracle"))
        ],
        exp.DataType.Type.NVARCHAR.value: [
            partial(anonymous, func="NVL(TRIM(TO_CHAR({})),'_null_recon_')", dialect=get_dialect("oracle"))
        ],
    },
    "databricks": {
        exp.DataType.Type.ARRAY.value: [
            partial(anonymous, func="CONCAT_WS(',', SORT_ARRAY({}))", dialect=get_dialect("databricks"))
        ],
    },
    "tsql": {
        "default": [partial(anonymous, func="COALESCE(LTRIM(RTRIM(CAST([{}] AS VARCHAR(256)))), '_null_recon_')")],
        exp.DataType.Type.DATE.value: [partial(anonymous, func="COALESCE(CONVERT(DATE, {0}, 101), '1900-01-01')")],
        exp.DataType.Type.TIME.value: [partial(anonymous, func="COALESCE(CONVERT(TIME, {0}, 108), '00:00:00')")],
        exp.DataType.Type.DATETIME.value: [
            partial(anonymous, func="COALESCE(CONVERT(DATETIME, {0}, 120), '1900-01-01 00:00:00')")
        ],
    },
}

sha256_partial = partial(sha2, num_bits="256", is_expr=True)
md5_partial = partial(md5, is_expr=True)
Dialect_hash_algo_mapping: dict[Dialect, HashAlgoMapping] = {
    get_dialect("snowflake"): HashAlgoMapping(
        source=sha256_partial,
        target=sha256_partial,
    ),
    get_dialect("oracle"): HashAlgoMapping(
        source=partial(
            anonymous, func="DBMS_CRYPTO.HASH(RAWTOHEX({}), 2)", is_expr=True, dialect=get_dialect("oracle")
        ),
        target=md5_partial,
    ),
    get_dialect("databricks"): HashAlgoMapping(
        source=sha256_partial,
        target=sha256_partial,
    ),
    get_dialect("tsql"): HashAlgoMapping(
        source=partial(
            anonymous, func="CONVERT(VARCHAR(256), HASHBYTES('SHA2_256', CONVERT(VARCHAR(256),{})), 2)", is_expr=True
        ),
        target=sha256_partial,
    ),
}



import logging

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.types import BooleanType, StringType, StructField, StructType
from sqlglot import Dialect, parse_one

from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_dialect
from databricks.labs.lakebridge.reconcile.recon_config import Schema, Table
from databricks.labs.lakebridge.reconcile.recon_output_config import SchemaMatchResult, SchemaReconcileOutput
from databricks.labs.lakebridge.transpiler.sqlglot.generator.databricks import Databricks

logger = logging.getLogger(__name__)


class SchemaCompare:
    def __init__(
        self,
        spark: SparkSession,
    ):
        self.spark = spark

    _schema_compare_output_schema: StructType = StructType(
        [
            StructField("source_column", StringType(), False),
            StructField("source_datatype", StringType(), False),
            StructField("databricks_column", StringType(), True),
            StructField("databricks_datatype", StringType(), True),
            StructField("is_valid", BooleanType(), False),
        ]
    )

    @classmethod
    def _build_master_schema(
        cls,
        source_schema: list[Schema],
        databricks_schema: list[Schema],
        table_conf: Table,
    ) -> list[SchemaMatchResult]:
        master_schema = source_schema
        if table_conf.select_columns:
            master_schema = [schema for schema in master_schema if schema.column_name in table_conf.select_columns]
        if table_conf.drop_columns:
            master_schema = [sschema for sschema in master_schema if sschema.column_name not in table_conf.drop_columns]

        target_column_map = table_conf.to_src_col_map or {}
        master_schema_match_res = [
            SchemaMatchResult(
                source_column_normalized=s.source_normalized_column_name,
                source_column_normalized_ansi=s.ansi_normalized_column_name,
                source_datatype=s.data_type,
                databricks_column=target_column_map.get(s.ansi_normalized_column_name, s.ansi_normalized_column_name),
                databricks_datatype=next(
                    (
                        tgt.data_type
                        for tgt in databricks_schema
                        if tgt.ansi_normalized_column_name
                        == target_column_map.get(s.ansi_normalized_column_name, s.ansi_normalized_column_name)
                    ),
                    "",
                ),
            )
            for s in master_schema
        ]
        return master_schema_match_res

    def _create_output_dataframe(self, data: list[SchemaMatchResult], schema: StructType) -> DataFrame:
        """Return a user-friendly dataframe for schema compare result."""
        transformed = []
        for item in data:
            output = tuple(
                [
                    DialectUtils.unnormalize_identifier(item.source_column_normalized_ansi),
                    item.source_datatype,
                    DialectUtils.unnormalize_identifier(item.databricks_column),
                    item.databricks_datatype,
                    item.is_valid,
                ]
            )
            transformed.append(output)

        return self.spark.createDataFrame(transformed, schema)

    @classmethod
    def _parse(cls, source: Dialect, column: str, data_type: str) -> str:
        return (
            parse_one(f"create table dummy ({column} {data_type})", read=source)
            .sql(dialect=get_dialect("databricks"))
            .replace(", ", ",")
        )

    @classmethod
    def _table_schema_status(cls, schema_compare_maps: list[SchemaMatchResult]) -> bool:
        return bool(all(x.is_valid for x in schema_compare_maps))

    @classmethod
    def _validate_parsed_query(cls, master: SchemaMatchResult, parsed_query) -> None:
        databricks_query = f"create table dummy ({master.source_column_normalized_ansi} {master.databricks_datatype})"
        logger.info(
            f"""
        Source datatype: create table dummy ({master.source_column_normalized} {master.source_datatype})
        Parse datatype: {parsed_query}
        Databricks datatype: {databricks_query}
        """
        )
        if parsed_query.lower() != databricks_query.lower():
            master.is_valid = False

    def compare(
        self,
        source_schema: list[Schema],
        databricks_schema: list[Schema],
        source: Dialect,
        table_conf: Table,
    ) -> SchemaReconcileOutput:
        """
        This method compares the source schema and the Databricks schema. It checks if the data types of the columns in the source schema
        match with the corresponding columns in the Databricks schema by parsing using remorph transpile.

        Returns:
            SchemaReconcileOutput: A dataclass object containing a boolean indicating the overall result of the comparison and a DataFrame with the comparison details.
        """
        master_schema = self._build_master_schema(source_schema, databricks_schema, table_conf)
        for master in master_schema:
            if not isinstance(source, Databricks):
                parsed_query = self._parse(source, master.source_column_normalized, master.source_datatype)
                self._validate_parsed_query(master, parsed_query)
            elif master.source_datatype.lower() != master.databricks_datatype.lower():
                master.is_valid = False

        df = self._create_output_dataframe(master_schema, self._schema_compare_output_schema)
        final_result = self._table_schema_status(master_schema)
        return SchemaReconcileOutput(final_result, df)



from collections.abc import Iterable

from databricks.labs.lakebridge.connections.database_manager import DatabaseManager
from databricks.labs.lakebridge.discovery.table import TableDefinition, TableFQN, FieldInfo
from databricks.labs.lakebridge.discovery.table_definition import TableDefinitionService


class TsqlTableDefinitionService(TableDefinitionService):

    # Hexadecimal value of § is U+00A7.Hexadecimal value of ‡ (double dagger) is U+2021
    def _get_table_definition_query(self, catalog_name: str) -> str:
        query = f"""
        WITH column_info AS (
            SELECT
                TABLE_CATALOG,
                TABLE_SCHEMA,
                TABLE_NAME,
                STRING_AGG(
                    CONCAT(
                        column_name,
                        '§',
                        CASE
                            WHEN numeric_precision IS NOT NULL AND numeric_scale IS NOT NULL THEN CONCAT(data_type, '(', numeric_precision, ',', numeric_scale, ')')
                            WHEN LOWER(data_type) = 'text' THEN CONCAT('varchar', '(', CHARACTER_MAXIMUM_LENGTH, ')')
                            ELSE data_type
                        END,
                        '§',
                        CASE
                            WHEN cis.IS_NULLABLE = 'YES' THEN 'true'
                            ELSE 'false'
                        END,
                        '§',
                        ISNULL(CAST(ep_col.value AS NVARCHAR(MAX)), '')
                    ),
                    '‡'
                ) WITHIN GROUP (ORDER BY ordinal_position) AS DERIVED_SCHEMA
            FROM
                {catalog_name}.sys.tables t
                INNER JOIN {catalog_name}.sys.columns c ON t.object_id = c.object_id
                INNER JOIN {catalog_name}.INFORMATION_SCHEMA.COLUMNS cis ON t.name = cis.TABLE_NAME AND c.name = cis.COLUMN_NAME
                OUTER APPLY (
                    SELECT TOP 1 value
                    FROM {catalog_name}.sys.extended_properties
                    WHERE major_id = t.object_id AND minor_id = 0
                    ORDER BY name DESC
                ) ep_tbl
                OUTER APPLY (
                    SELECT TOP 1 value
                    FROM {catalog_name}.sys.extended_properties
                    WHERE major_id = c.object_id AND minor_id = c.column_id
                    ORDER BY name DESC
                ) ep_col
            GROUP BY
                TABLE_CATALOG,
                TABLE_SCHEMA,
                TABLE_NAME
        ),
        table_file_info AS (
            SELECT
                s.name AS TABLE_SCHEMA,
                t.name AS TABLE_NAME,
                f.physical_name AS location,
                f.type_desc AS TABLE_FORMAT,
                CAST(ROUND(SUM(a.used_pages) * 8.0 / 1024, 2) AS DECIMAL(18, 2)) AS SIZE_GB
            FROM
                {catalog_name}.sys.tables t
                INNER JOIN {catalog_name}.sys.indexes i ON t.object_id = i.object_id
                INNER JOIN {catalog_name}.sys.partitions p ON i.object_id = p.object_id AND i.index_id = p.index_id
                INNER JOIN {catalog_name}.sys.allocation_units a ON p.partition_id = a.container_id
                INNER JOIN {catalog_name}.sys.schemas s ON t.schema_id = s.schema_id
                INNER JOIN {catalog_name}.sys.database_files f ON a.data_space_id = f.data_space_id
                LEFT JOIN {catalog_name}.sys.extended_properties ep ON ep.major_id = t.object_id AND ep.minor_id = 0
            GROUP BY
                s.name,
                t.name,
                f.name,
                f.physical_name,
                f.type_desc
        ),
        table_comment_info AS (
            SELECT
                s.name AS TABLE_SCHEMA,
                t.name AS TABLE_NAME,
                CAST(ep.value AS NVARCHAR(MAX)) AS TABLE_COMMENT
            FROM
                {catalog_name}.sys.tables t
                INNER JOIN {catalog_name}.sys.schemas s ON t.schema_id = s.schema_id
                OUTER APPLY (
                    SELECT TOP 1 value
                    FROM {catalog_name}.sys.extended_properties
                    WHERE major_id = t.object_id AND minor_id = 0
                    ORDER BY name DESC
                ) ep
        ),
        table_pk_info AS (
            SELECT
                TC.TABLE_CATALOG,
                TC.TABLE_SCHEMA,
                TC.TABLE_NAME,
                STRING_AGG(KU.COLUMN_NAME,':') as PK_COLUMN_NAME
                FROM {catalog_name}.INFORMATION_SCHEMA.TABLE_CONSTRAINTS AS TC
                JOIN {catalog_name}.INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS KU
                ON TC.CONSTRAINT_NAME = KU.CONSTRAINT_NAME
                AND TC.TABLE_NAME = KU.TABLE_NAME
                WHERE TC.CONSTRAINT_TYPE = 'PRIMARY KEY' group by TC.TABLE_CATALOG, TC.TABLE_SCHEMA, TC.TABLE_NAME)
        SELECT
            sft.TABLE_CATALOG,
            sft.TABLE_SCHEMA,
            sft.TABLE_NAME,
            tfi.location,
            tfi.TABLE_FORMAT,
            '' as view_definition,
            column_info.DERIVED_SCHEMA,
            tfi.SIZE_GB,
            tci.TABLE_COMMENT,
            tpK.PK_COLUMN_NAME
        FROM
            column_info
            JOIN {catalog_name}.INFORMATION_SCHEMA.TABLES sft ON column_info.TABLE_CATALOG = sft.TABLE_CATALOG AND column_info.TABLE_SCHEMA = sft.TABLE_SCHEMA AND column_info.TABLE_NAME = sft.TABLE_NAME
            LEFT JOIN table_file_info tfi ON column_info.TABLE_SCHEMA = tfi.TABLE_SCHEMA AND column_info.TABLE_NAME = tfi.TABLE_NAME
            LEFT JOIN table_comment_info tci ON column_info.TABLE_SCHEMA = tci.TABLE_SCHEMA AND column_info.TABLE_NAME = tci.TABLE_NAME
            LEFT JOIN table_pk_info tpK ON column_info.TABLE_SCHEMA = tpK.TABLE_SCHEMA AND column_info.TABLE_NAME = tpK.TABLE_NAME

        UNION ALL
        SELECT
            sfv.TABLE_CATALOG,
            sfv.TABLE_SCHEMA,
            sfv.TABLE_NAME,
            '' location,
            '' TABLE_FORMAT,
            sfv.view_definition,
            '' DERIVED_SCHEMA,
            0 SIZE_GB,
            '' TABLE_COMMENT,
            '' PK_COLUMN_NAME
        FROM {catalog_name}.INFORMATION_SCHEMA.VIEWS sfv
        """
        return query

    def get_table_definition(self, catalog_name: str) -> Iterable[TableDefinition]:
        sql = self._get_table_definition_query(catalog_name)
        tsql_connection = self.connection
        result = tsql_connection.execute_query(sql)

        column_names = list(result.keys())
        table_definitions = []

        for row in result:
            result = dict(zip(column_names, row))
            table_fqn = TableFQN(
                catalog=result["TABLE_CATALOG"], schema=result["TABLE_SCHEMA"], name=result["TABLE_NAME"]
            )
            columns = result["DERIVED_SCHEMA"].split("‡") if result["DERIVED_SCHEMA"] else None
            field_info = []
            if columns is not None:
                for column in columns:
                    column_info = column.split("§")
                    field = FieldInfo(
                        name=column_info[0],
                        data_type=column_info[1],
                        nullable=column_info[2],
                        comment=column_info[3],
                    )
                    field_info.append(field)

            pks = result["PK_COLUMN_NAME"].split(":") if result["PK_COLUMN_NAME"] else None
            table_definition = TableDefinition(
                fqn=table_fqn,
                location=result["location"],
                table_format=result["TABLE_FORMAT"],
                view_text=result["view_definition"],
                columns=field_info,
                size_gb=result["SIZE_GB"],
                comment=result["TABLE_COMMENT"],
                primary_keys=pks,
            )
            table_definitions.append(table_definition)
        return table_definitions

    def get_all_catalog(self) -> Iterable[str]:
        cursor: DatabaseManager = self.connection
        result = cursor.connector.execute_query("""select name from sys.databases""")
        catalogs = [row[0] for row in result]
        print(catalogs)
        return catalogs



from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any

from databricks.labs.lakebridge.discovery.table import TableDefinition


class TableDefinitionService(ABC):

    def __init__(self, connection: Any):
        self.connection = connection

    @abstractmethod
    def get_table_definition(self, catalog_name: str) -> Iterable[TableDefinition]:
        pass

    @abstractmethod
    def _get_table_definition_query(self, catalog_name: str) -> str:
        pass

    @abstractmethod
    def get_all_catalog(self) -> Iterable[str]:
        pass



import logging
from sqlglot.dialects.presto import Presto as presto
from sqlglot import exp
from sqlglot.helper import seq_get
from sqlglot.errors import ParseError
from sqlglot.tokens import TokenType

from databricks.labs.lakebridge.transpiler.sqlglot import local_expression


logger = logging.getLogger(__name__)


def _build_approx_percentile(args: list) -> exp.Expression:
    if len(args) == 4:
        arg3 = seq_get(args, 3)
        try:
            number = float(arg3.this) if arg3 is not None else 0
            return exp.ApproxQuantile(
                this=seq_get(args, 0),
                weight=seq_get(args, 1),
                quantile=seq_get(args, 2),
                accuracy=exp.Literal(this=f'{int((1/number) * 100)} ', is_string=False),
            )
        except ValueError as exc:
            raise ParseError(f"Expected a string representation of a number for argument 2, but got {arg3}") from exc
    if len(args) == 3:
        arg2 = seq_get(args, 2)
        try:
            number = float(arg2.this) if arg2 is not None else 0
            return exp.ApproxQuantile(
                this=seq_get(args, 0),
                quantile=seq_get(args, 1),
                accuracy=exp.Literal(this=f'{int((1/number) * 100)}', is_string=False),
            )
        except ValueError as exc:
            raise ParseError(f"Expected a string representation of a number for argument 2, but got {arg2}") from exc
    return exp.ApproxQuantile.from_arg_list(args)


def _build_any_keys_match(args: list) -> local_expression.ArrayExists:
    return local_expression.ArrayExists(
        this=local_expression.MapKeys(this=seq_get(args, 0)), expression=seq_get(args, 1)
    )


def _build_str_position(args: list) -> local_expression.Locate:
    # TODO the 3rd param in presto strpos and databricks locate has different implementation.
    # For now we haven't implemented the logic same as presto for 3rd param.
    # Users should be vigilant when using 3 param function in presto strpos.
    if len(args) == 3:
        msg = (
            "*Warning:: The third parameter in Presto's `strpos` function and Databricks' `locate` function "
            "have different implementations. Please exercise caution when using the three-parameter version "
            "of the `strpos` function in Presto."
        )
        logger.warning(msg)
        return local_expression.Locate(substring=seq_get(args, 1), this=seq_get(args, 0), position=seq_get(args, 2))
    return local_expression.Locate(substring=seq_get(args, 1), this=seq_get(args, 0))


def _build_array_average(args: list) -> exp.Reduce:
    return exp.Reduce(
        this=exp.ArrayFilter(
            this=seq_get(args, 0),
            expression=exp.Lambda(
                this=exp.Not(this=exp.Is(this=exp.Identifier(this="x", quoted=False), expression=exp.Null())),
                expressions=[exp.Identifier(this="x", quoted=False)],
            ),
        ),
        initial=local_expression.NamedStruct(
            expressions=[
                exp.Literal(this="sum", is_string=True),
                exp.Cast(this=exp.Literal(this="0", is_string=False), to=exp.DataType(this="DOUBLE")),
                exp.Literal(this="cnt", is_string=True),
                exp.Literal(this="0", is_string=False),
            ],
        ),
        merge=exp.Lambda(
            this=local_expression.NamedStruct(
                expressions=[
                    exp.Literal(this="sum", is_string=True),
                    exp.Add(
                        this=exp.Dot(
                            this=exp.Identifier(this="acc", quoted=False),
                            expression=exp.Identifier(this="sum", quoted=False),
                        ),
                        expression=exp.Identifier(this="x", quoted=False),
                    ),
                    exp.Literal(this="cnt", is_string=True),
                    exp.Add(
                        this=exp.Dot(
                            this=exp.Identifier(this="acc", quoted=False),
                            expression=exp.Identifier(this="cnt", quoted=False),
                        ),
                        expression=exp.Literal(this="1", is_string=False),
                    ),
                ],
            ),
            expressions=[exp.Identifier(this="acc", quoted=False), exp.Identifier(this="x", quoted=False)],
        ),
        finish=exp.Lambda(
            this=exp.Anonymous(
                this="try_divide",
                expressions=[
                    exp.Dot(
                        this=exp.Identifier(this="acc", quoted=False),
                        expression=exp.Identifier(this="sum", quoted=False),
                    ),
                    exp.Dot(
                        this=exp.Identifier(this="acc", quoted=False),
                        expression=exp.Identifier(this="cnt", quoted=False),
                    ),
                ],
            ),
            expressions=[exp.Identifier(this="acc", quoted=False)],
        ),
    )


def _build_json_size(args: list):
    return exp.Case(
        ifs=[
            exp.If(
                this=exp.Like(
                    this=local_expression.GetJsonObject(
                        this=exp.Column(this=seq_get(args, 0)),
                        path=exp.Column(this=seq_get(args, 1)),
                    ),
                    expression=exp.Literal(this="{%", is_string=True),
                ),
                true=exp.ArraySize(
                    this=exp.Anonymous(
                        this="from_json",
                        expressions=[
                            local_expression.GetJsonObject(
                                this=exp.Column(this=seq_get(args, 0)),
                                path=exp.Column(this=seq_get(args, 1)),
                            ),
                            exp.Literal(this="map<string,string>", is_string=True),
                        ],
                    )
                ),
            ),
            exp.If(
                this=exp.Like(
                    this=local_expression.GetJsonObject(
                        this=exp.Column(this=seq_get(args, 0)),
                        path=exp.Column(this=seq_get(args, 1)),
                    ),
                    expression=exp.Literal(this="[%", is_string=True),
                ),
                true=exp.ArraySize(
                    this=exp.Anonymous(
                        this="from_json",
                        expressions=[
                            local_expression.GetJsonObject(
                                this=exp.Column(this=seq_get(args, 0)),
                                path=exp.Column(this=seq_get(args, 1)),
                            ),
                            exp.Literal(this="array<string>", is_string=True),
                        ],
                    )
                ),
            ),
            exp.If(
                this=exp.Not(
                    this=exp.Is(
                        this=local_expression.GetJsonObject(
                            this=exp.Column(this=seq_get(args, 0)),
                            path=exp.Column(this=seq_get(args, 1)),
                        ),
                        expression=exp.Null(),
                    )
                ),
                true=exp.Literal(this="0", is_string=False),
            ),
        ],
        default=exp.Null(),
    )


class Presto(presto):

    class Parser(presto.Parser):
        VALUES_FOLLOWED_BY_PAREN = False

        FUNCTIONS = {
            **presto.Parser.FUNCTIONS,
            "APPROX_PERCENTILE": _build_approx_percentile,
            "STRPOS": _build_str_position,
            "ANY_KEYS_MATCH": _build_any_keys_match,
            "ARRAY_AVERAGE": _build_array_average,
            "JSON_SIZE": _build_json_size,
            "FORMAT_DATETIME": local_expression.DateFormat.from_arg_list,
        }

    class Tokenizer(presto.Tokenizer):
        KEYWORDS = {
            **presto.Tokenizer.KEYWORDS,
            "JSON": TokenType.TEXT,
        }



from pathlib import Path
from subprocess import run, CalledProcessError
from dataclasses import dataclass
from enum import Enum

import venv
import tempfile
import json
import logging
import yaml
import duckdb

from databricks.labs.lakebridge.connections.credential_manager import cred_file

from databricks.labs.lakebridge.assessments.profiler_config import PipelineConfig, Step
from databricks.labs.lakebridge.connections.database_manager import DatabaseManager

logger = logging.getLogger(__name__)
logger.setLevel("INFO")

DB_NAME = "profiler_extract.db"


class StepExecutionStatus(str, Enum):
    COMPLETE = "COMPLETE"
    ERROR = "ERROR"
    SKIPPED = "SKIPPED"


@dataclass
class StepExecutionResult:
    step_name: str
    status: StepExecutionStatus
    error_message: str | None = None


class PipelineClass:
    def __init__(self, config: PipelineConfig, executor: DatabaseManager):
        self.config = config
        self.executor = executor
        self.db_path_prefix = Path(config.extract_folder)

    def execute(self) -> list[StepExecutionResult]:
        logging.info(f"Pipeline initialized with config: {self.config.name}, version: {self.config.version}")
        execution_results: list[StepExecutionResult] = []
        for step in self.config.steps:
            result = self._process_step(step)
            execution_results.append(result)
            logging.info(f"Step '{step.name}' completed with status: {result.status}")

        logging.info("Pipeline execution completed")
        return execution_results

    def _process_step(self, step: Step) -> StepExecutionResult:
        if step.flag != "active":
            logging.info(f"Skipping step: {step.name} as it is not active")
            return StepExecutionResult(step_name=step.name, status=StepExecutionStatus.SKIPPED)

        logging.debug(f"Executing step: {step.name}")
        try:
            status = self._execute_step(step)
            return StepExecutionResult(step_name=step.name, status=status)
        except RuntimeError as e:
            return StepExecutionResult(step_name=step.name, status=StepExecutionStatus.ERROR, error_message=str(e))

    def _execute_step(self, step: Step) -> StepExecutionStatus:
        if step.type == "sql":
            logging.info(f"Executing SQL step {step.name}")
            self._execute_sql_step(step)
            return StepExecutionStatus.COMPLETE
        if step.type == "python":
            logging.info(f"Executing Python step {step.name}")
            self._execute_python_step(step)
            return StepExecutionStatus.COMPLETE
        logging.error(f"Unsupported step type: {step.type}")
        raise RuntimeError(f"Unsupported step type: {step.type}")

    def _execute_sql_step(self, step: Step):
        logging.debug(f"Reading query from file: {step.extract_source}")
        with open(step.extract_source, 'r', encoding='utf-8') as file:
            query = file.read()

        # Execute the query using the database manager
        logging.info(f"Executing query: {query}")
        try:
            result = self.executor.execute_query(query)

            # Save the result to duckdb
            self._save_to_db(result, step.name, str(step.mode))
        except Exception as e:
            logging.error(f"SQL execution failed: {str(e)}")
            raise RuntimeError(f"SQL execution failed: {str(e)}") from e

    def _execute_python_step(self, step: Step):

        logging.debug(f"Executing Python script: {step.extract_source}")
        db_path = str(self.db_path_prefix / DB_NAME)
        credential_config = str(cred_file("lakebridge"))

        # Create a temporary directory for the virtual environment
        with tempfile.TemporaryDirectory() as temp_dir:
            venv_dir = Path(temp_dir) / "venv"
            venv.create(venv_dir, with_pip=True)
            venv_python = venv_dir / "bin" / "python"
            venv_pip = venv_dir / "bin" / "pip"

            logger.info(f"Creating a virtual environment for Python script execution: ${venv_dir}")
            # Install dependencies in the virtual environment
            if step.dependencies:
                logging.info(f"Installing dependencies: {', '.join(step.dependencies)}")
                try:
                    logging.debug("Upgrading local pip")
                    run([str(venv_pip), "install", "--upgrade", "pip"], check=True, capture_output=True, text=True)

                    run([str(venv_pip), "install", *step.dependencies], check=True, capture_output=True, text=True)
                except CalledProcessError as e:
                    logging.error(f"Failed to install dependencies: {e.stderr}")
                    raise RuntimeError(f"Failed to install dependencies: {e.stderr}") from e

            # Execute the Python script using the virtual environment's Python interpreter
            try:
                result = run(
                    [
                        str(venv_python),
                        str(step.extract_source),
                        "--db-path",
                        db_path,
                        "--credential-config-path",
                        credential_config,
                    ],
                    check=True,
                    capture_output=True,
                    text=True,
                )

                try:
                    output = json.loads(result.stdout)
                    if output["status"] == "success":
                        logging.info(f"Python script completed: {output['message']}")
                    else:
                        raise RuntimeError(f"Script reported error: {output['message']}")
                except json.JSONDecodeError:
                    logging.info(f"Python script output: {result.stdout}")

            except CalledProcessError as e:
                error_msg = e.stderr
                logging.error(f"Python script failed: {error_msg}")
                raise RuntimeError(f"Script execution failed: {error_msg}") from e

    def _save_to_db(self, result, step_name: str, mode: str, batch_size: int = 1000):
        self._create_dir(self.db_path_prefix)
        db_path = str(self.db_path_prefix / DB_NAME)

        with duckdb.connect(db_path) as conn:
            columns = result.keys()
            # TODO: Add support for figuring out data types from SQLALCHEMY result object result.cursor.description is not reliable
            schema = ' STRING, '.join(columns) + ' STRING'

            # Handle write modes
            if mode == 'overwrite':
                conn.execute(f"CREATE OR REPLACE TABLE {step_name} ({schema})")
            elif mode == 'append' and step_name not in conn.get_table_names(""):
                conn.execute(f"CREATE TABLE {step_name} ({schema})")

            # Batch insert using prepared statements
            placeholders = ', '.join(['?' for _ in columns])
            insert_query = f"INSERT INTO {step_name} VALUES ({placeholders})"

            # Fetch and insert rows in batches
            while True:
                rows = result.fetchmany(batch_size)
                if not rows:
                    break
                conn.executemany(insert_query, rows)

    @staticmethod
    def _create_dir(dir_path: Path):
        if not Path(dir_path).exists():
            dir_path.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def load_config_from_yaml(file_path: str) -> PipelineConfig:
        with open(file_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
        steps = [Step(**step) for step in data['steps']]
        return PipelineConfig(
            name=data['name'], version=data['version'], extract_folder=data['extract_folder'], steps=steps
        )



import logging
import webbrowser

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installation import SerdeError
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, PermissionDenied

from databricks.labs.lakebridge.config import ReconcileConfig, TableRecon
from databricks.labs.lakebridge.deployment.recon import RECON_JOB_NAME
from databricks.labs.lakebridge.reconcile.recon_config import RECONCILE_OPERATION_NAME

logger = logging.getLogger(__name__)

_RECON_README_URL = "https://github.com/databrickslabs/lakebridge/blob/main/docs/recon_configurations/README.md"


class ReconcileRunner:
    def __init__(
        self,
        ws: WorkspaceClient,
        installation: Installation,
        install_state: InstallState,
        prompts: Prompts,
    ):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state
        self._prompts = prompts

    def run(self, operation_name=RECONCILE_OPERATION_NAME):
        reconcile_config = self._get_verified_recon_config()
        job_id = self._get_recon_job_id(reconcile_config)
        logger.info(f"Triggering the reconcile job with job_id: `{job_id}`")
        wait = self._ws.jobs.run_now(job_id, job_parameters={"operation_name": operation_name})
        if not wait.run_id:
            raise SystemExit(f"Job {job_id} execution failed. Please check the job logs for more details.")

        job_run_url = f"{self._ws.config.host}/jobs/{job_id}/runs/{wait.run_id}"
        logger.info(
            f"'{operation_name.upper()}' job started. Please check the job_url `{job_run_url}` for the current status."
        )
        if self._prompts.confirm(f"Would you like to open the job run URL `{job_run_url}` in the browser?"):
            webbrowser.open(job_run_url)

    def _get_verified_recon_config(self) -> ReconcileConfig:
        try:
            recon_config = self._installation.load(ReconcileConfig)
        except NotFound as err:
            raise SystemExit("Cannot find existing `reconcile` installation. Please try reinstalling.") from err
        except (PermissionDenied, SerdeError, ValueError, AttributeError) as e:
            install_dir = self._installation.install_folder()
            raise SystemExit(
                f"Existing `reconcile` installation at {install_dir} is corrupted. Please try reinstalling."
            ) from e

        self._verify_recon_table_config(recon_config)
        return recon_config

    def _verify_recon_table_config(self, recon_config):
        source_catalog_or_schema = (
            recon_config.database_config.source_catalog
            if recon_config.database_config.source_catalog
            else recon_config.database_config.source_schema
        )
        # Filename pattern for recon table config `recon_config_<SOURCE>_<CATALOG_OR_SCHEMA>_<FILTER_TYPE>.json`
        # Example: recon_config_snowflake_sample_data_all.json
        filename = f"recon_config_{recon_config.data_source}_{source_catalog_or_schema}_{recon_config.report_type}.json"
        try:
            logger.debug(f"Loading recon table config `{filename}` from workspace.")
            self._installation.load(TableRecon, filename=filename)
        except NotFound as e:
            err_msg = (
                "Cannot find recon table configuration in existing `reconcile` installation. "
                f"Please provide the configuration file {filename} in the workspace."
            )
            logger.error(f"{err_msg}. For more details, please refer to {_RECON_README_URL}")
            raise SystemExit(err_msg) from e
        except (PermissionDenied, SerdeError, ValueError, AttributeError) as e:
            install_dir = self._installation.install_folder()
            err_msg = (
                f"Cannot load corrupted recon table configuration from {install_dir}/{filename}. "
                f"Please validate the file."
            )
            logger.error(f"{err_msg}. For more details, please refer to {_RECON_README_URL}")
            raise SystemExit(err_msg) from e

    def _get_recon_job_id(self, reconcile_config: ReconcileConfig) -> int:
        if reconcile_config.job_id:
            logger.debug("Reconcile job id found in the reconcile config.")
            return int(reconcile_config.job_id)
        if RECON_JOB_NAME in self._install_state.jobs:
            logger.debug("Reconcile job id found in the install state.")
            return int(self._install_state.jobs[RECON_JOB_NAME])
        raise SystemExit("Reconcile Job ID not found. Please try reinstalling.")



import dataclasses
import logging
import os
import webbrowser
from collections.abc import Set, Callable, Sequence
from pathlib import Path
from typing import Any, cast

from databricks.labs.blueprint.installation import Installation, JsonValue, SerdeError
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.wheels import ProductInfo
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound, PermissionDenied

from databricks.labs.lakebridge.__about__ import __version__
from databricks.labs.lakebridge.config import (
    DatabaseConfig,
    ReconcileConfig,
    LakebridgeConfiguration,
    ReconcileMetadataConfig,
    TranspileConfig,
)
from databricks.labs.lakebridge.contexts.application import ApplicationContext
from databricks.labs.lakebridge.deployment.configurator import ResourceConfigurator
from databricks.labs.lakebridge.deployment.installation import WorkspaceInstallation
from databricks.labs.lakebridge.reconcile.constants import ReconReportType, ReconSourceType
from databricks.labs.lakebridge.transpiler.installers import (
    BladebridgeInstaller,
    MorpheusInstaller,
    TranspilerInstaller,
)
from databricks.labs.lakebridge.transpiler.repository import TranspilerRepository

logger = logging.getLogger(__name__)

TRANSPILER_WAREHOUSE_PREFIX = "Lakebridge Transpiler Validation"


class WorkspaceInstaller:
    def __init__(
        self,
        ws: WorkspaceClient,
        prompts: Prompts,
        installation: Installation,
        install_state: InstallState,
        product_info: ProductInfo,
        resource_configurator: ResourceConfigurator,
        workspace_installation: WorkspaceInstallation,
        environ: dict[str, str] | None = None,
        *,
        transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
        transpiler_installers: Sequence[Callable[[TranspilerRepository], TranspilerInstaller]] = (
            BladebridgeInstaller,
            MorpheusInstaller,
        ),
    ):
        self._ws = ws
        self._prompts = prompts
        self._installation = installation
        self._install_state = install_state
        self._product_info = product_info
        self._resource_configurator = resource_configurator
        self._ws_installation = workspace_installation
        self._transpiler_repository = transpiler_repository
        self._transpiler_installer_factories = transpiler_installers

        if not environ:
            environ = dict(os.environ.items())

        if "DATABRICKS_RUNTIME_VERSION" in environ:
            msg = "WorkspaceInstaller is not supposed to be executed in Databricks Runtime"
            raise SystemExit(msg)

    @property
    def _transpiler_installers(self) -> Set[TranspilerInstaller]:
        return frozenset(factory(self._transpiler_repository) for factory in self._transpiler_installer_factories)

    def run(
        self, module: str, config: LakebridgeConfiguration | None = None, artifact: str | None = None
    ) -> LakebridgeConfiguration:
        logger.debug(f"Initializing workspace installation for module: {module} (config: {config})")
        if module == "transpile" and artifact:
            self._install_artifact(artifact)
        elif module in {"transpile", "all"}:
            for transpiler_installer in self._transpiler_installers:
                transpiler_installer.install()
        if not config:
            config = self.configure(module)
        if self._is_testing():
            return config
        self._ws_installation.install(config)
        logger.info("Installation completed successfully! Please refer to the documentation for the next steps.")
        return config

    def upgrade_installed_transpilers(self) -> bool:
        """Detect and upgrade, if possible and necessary, installed transpilers."""
        installed_transpilers = self._transpiler_repository.all_transpiler_names()
        if installed_transpilers:
            logger.info(f"Detected installed transpilers: {sorted(installed_transpilers)}")
        upgraded = False
        for transpiler_installer in self._transpiler_installers:
            name = transpiler_installer.name
            if name in installed_transpilers:
                logger.info(f"Checking for {name} upgrades...")
                upgraded |= transpiler_installer.install()
        # If we upgraded anything, the configuration process needs to run again.
        if upgraded:
            config = self.configure("transpile")
            if not self._is_testing():
                self._ws_installation.install(config)
        return upgraded

    def _install_artifact(self, artifact: str) -> None:
        path = Path(artifact)
        if not path.exists():
            logger.error(f"Could not locate artifact {artifact}")
            return
        for transpiler_installer in self._transpiler_installers:
            if transpiler_installer.can_install(path):
                transpiler_installer.install(path)
                break
        else:
            logger.fatal(f"Cannot install unsupported artifact: {artifact}")

    def configure(self, module: str) -> LakebridgeConfiguration:
        match module:
            case "transpile":
                logger.info("Configuring lakebridge `transpile`.")
                return LakebridgeConfiguration(self._configure_transpile(), None)
            case "reconcile":
                logger.info("Configuring lakebridge `reconcile`.")
                return LakebridgeConfiguration(None, self._configure_reconcile())
            case "all":
                logger.info("Configuring lakebridge `transpile` and `reconcile`.")
                return LakebridgeConfiguration(
                    self._configure_transpile(),
                    self._configure_reconcile(),
                )
            case _:
                raise ValueError(f"Invalid input: {module}")

    def _is_testing(self):
        return self._product_info.product_name() != "lakebridge"

    def _configure_transpile(self) -> TranspileConfig:
        try:
            config = self._installation.load(TranspileConfig)
            logger.info("Lakebridge `transpile` is already installed on this workspace.")
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                return config
        except NotFound:
            logger.info("Couldn't find existing `transpile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `transpile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        config = self._configure_new_transpile_installation()
        logger.info("Finished configuring lakebridge `transpile`.")
        return config

    def _configure_new_transpile_installation(self) -> TranspileConfig:
        default_config = self._prompt_for_new_transpile_installation()
        runtime_config = None
        catalog_name = "remorph"
        schema_name = "transpiler"
        if not default_config.skip_validation:
            catalog_name = self._configure_catalog()
            schema_name = self._configure_schema(catalog_name, "transpile")
            self._has_necessary_access(catalog_name, schema_name)
            warehouse_id = self._resource_configurator.prompt_for_warehouse_setup(TRANSPILER_WAREHOUSE_PREFIX)
            runtime_config = {"warehouse_id": warehouse_id}

        config = dataclasses.replace(
            default_config,
            catalog_name=catalog_name,
            schema_name=schema_name,
            sdk_config=runtime_config,
        )
        self._save_config(config)
        return config

    def _all_installed_dialects(self) -> list[str]:
        return sorted(self._transpiler_repository.all_dialects())

    def _transpilers_with_dialect(self, dialect: str) -> list[str]:
        return sorted(self._transpiler_repository.transpilers_with_dialect(dialect))

    def _transpiler_config_path(self, transpiler: str) -> Path:
        return self._transpiler_repository.transpiler_config_path(transpiler)

    def _prompt_for_new_transpile_installation(self) -> TranspileConfig:
        install_later = "Set it later"
        # TODO tidy this up, logger might not display the below in console...
        logger.info("Please answer a few questions to configure lakebridge `transpile`")
        all_dialects = [install_later, *self._all_installed_dialects()]
        source_dialect: str | None = self._prompts.choice("Select the source dialect:", all_dialects, sort=False)
        if source_dialect == install_later:
            source_dialect = None
        transpiler_name: str | None = None
        transpiler_config_path: Path | None = None
        if source_dialect:
            transpilers = self._transpilers_with_dialect(source_dialect)
            if len(transpilers) > 1:
                transpilers = [install_later] + transpilers
                transpiler_name = self._prompts.choice("Select the transpiler:", transpilers, sort=False)
                if transpiler_name == install_later:
                    transpiler_name = None
            else:
                transpiler_name = next(t for t in transpilers)
                logger.info(f"Lakebridge will use the {transpiler_name} transpiler")
            if transpiler_name:
                transpiler_config_path = self._transpiler_config_path(transpiler_name)
        transpiler_options: dict[str, JsonValue] | None = None
        if transpiler_config_path:
            transpiler_options = self._prompt_for_transpiler_options(
                cast(str, transpiler_name), cast(str, source_dialect)
            )
        input_source: str | None = self._prompts.question(
            "Enter input SQL path (directory/file)", default=install_later
        )
        if input_source == install_later:
            input_source = None
        output_folder = self._prompts.question("Enter output directory", default="transpiled")
        # When defaults are passed along we need to use absolute paths to avoid issues with relative paths
        if output_folder == "transpiled":
            output_folder = str(Path.cwd() / "transpiled")
        error_file_path = self._prompts.question("Enter error file path", default="errors.log")
        if error_file_path == "errors.log":
            error_file_path = str(Path.cwd() / "errors.log")

        run_validation = self._prompts.confirm(
            "Would you like to validate the syntax and semantics of the transpiled queries?"
        )

        return TranspileConfig(
            transpiler_config_path=str(transpiler_config_path) if transpiler_config_path is not None else None,
            transpiler_options=transpiler_options,
            source_dialect=source_dialect,
            skip_validation=(not run_validation),
            input_source=input_source,
            output_folder=output_folder,
            error_file_path=error_file_path,
        )

    def _prompt_for_transpiler_options(self, transpiler_name: str, source_dialect: str) -> dict[str, Any] | None:
        config_options = self._transpiler_repository.transpiler_config_options(transpiler_name, source_dialect)
        if len(config_options) == 0:
            return None
        return {option.flag: option.prompt_for_value(self._prompts) for option in config_options}

    def _configure_catalog(self) -> str:
        return self._resource_configurator.prompt_for_catalog_setup()

    def _configure_schema(
        self,
        catalog: str,
        default_schema_name: str,
    ) -> str:
        return self._resource_configurator.prompt_for_schema_setup(
            catalog,
            default_schema_name,
        )

    def _configure_reconcile(self) -> ReconcileConfig:
        try:
            self._installation.load(ReconcileConfig)
            logger.info("Lakebridge `reconcile` is already installed on this workspace.")
            if not self._prompts.confirm("Do you want to override the existing installation?"):
                # TODO: Exit gracefully, without raising SystemExit
                raise SystemExit(
                    "Lakebridge `reconcile` is already installed and no override has been requested. Exiting..."
                )
        except NotFound:
            logger.info("Couldn't find existing `reconcile` installation")
        except (PermissionDenied, SerdeError, ValueError, AttributeError):
            install_dir = self._installation.install_folder()
            logger.warning(
                f"Existing `reconcile` installation at {install_dir} is corrupted. Continuing new installation..."
            )

        config = self._configure_new_reconcile_installation()
        logger.info("Finished configuring lakebridge `reconcile`.")
        return config

    def _configure_new_reconcile_installation(self) -> ReconcileConfig:
        default_config = self._prompt_for_new_reconcile_installation()
        self._save_config(default_config)
        return default_config

    def _prompt_for_new_reconcile_installation(self) -> ReconcileConfig:
        logger.info("Please answer a few questions to configure lakebridge `reconcile`")
        data_source = self._prompts.choice(
            "Select the Data Source:", [source_type.value for source_type in ReconSourceType]
        )
        report_type = self._prompts.choice(
            "Select the report type:", [report_type.value for report_type in ReconReportType]
        )
        scope_name = self._prompts.question(
            f"Enter Secret scope name to store `{data_source.capitalize()}` connection details / secrets",
            default=f"remorph_{data_source}",
        )

        db_config = self._prompt_for_reconcile_database_config(data_source)
        metadata_config = self._prompt_for_reconcile_metadata_config()

        return ReconcileConfig(
            data_source=data_source,
            report_type=report_type,
            secret_scope=scope_name,
            database_config=db_config,
            metadata_config=metadata_config,
        )

    def _prompt_for_reconcile_database_config(self, source) -> DatabaseConfig:
        source_catalog = None
        if source == ReconSourceType.SNOWFLAKE.value:
            source_catalog = self._prompts.question(f"Enter source catalog name for `{source.capitalize()}`")

        schema_prompt = f"Enter source schema name for `{source.capitalize()}`"
        if source in {ReconSourceType.ORACLE.value}:
            schema_prompt = f"Enter source database name for `{source.capitalize()}`"

        source_schema = self._prompts.question(schema_prompt)
        target_catalog = self._prompts.question("Enter target catalog name for Databricks")
        target_schema = self._prompts.question("Enter target schema name for Databricks")

        return DatabaseConfig(
            source_schema=source_schema,
            target_catalog=target_catalog,
            target_schema=target_schema,
            source_catalog=source_catalog,
        )

    def _prompt_for_reconcile_metadata_config(self) -> ReconcileMetadataConfig:
        logger.info("Configuring reconcile metadata.")
        catalog = self._configure_catalog()
        schema = self._configure_schema(
            catalog,
            "reconcile",
        )
        volume = self._configure_volume(catalog, schema, "reconcile_volume")
        self._has_necessary_access(catalog, schema, volume)
        return ReconcileMetadataConfig(catalog=catalog, schema=schema, volume=volume)

    def _configure_volume(
        self,
        catalog: str,
        schema: str,
        default_volume_name: str,
    ) -> str:
        return self._resource_configurator.prompt_for_volume_setup(
            catalog,
            schema,
            default_volume_name,
        )

    def _save_config(self, config: TranspileConfig | ReconcileConfig):
        logger.info(f"Saving configuration file {config.__file__}")
        self._installation.save(config)
        ws_file_url = self._installation.workspace_link(config.__file__)
        if self._prompts.confirm(f"Open config file {ws_file_url} in the browser?"):
            webbrowser.open(ws_file_url)

    def _has_necessary_access(self, catalog_name: str, schema_name: str, volume_name: str | None = None):
        self._resource_configurator.has_necessary_access(catalog_name, schema_name, volume_name)


def installer(ws: WorkspaceClient, transpiler_repository: TranspilerRepository) -> WorkspaceInstaller:
    app_context = ApplicationContext(_verify_workspace_client(ws))
    return WorkspaceInstaller(
        app_context.workspace_client,
        app_context.prompts,
        app_context.installation,
        app_context.install_state,
        app_context.product_info,
        app_context.resource_configurator,
        app_context.workspace_installation,
        transpiler_repository=transpiler_repository,
    )


def _verify_workspace_client(ws: WorkspaceClient) -> WorkspaceClient:
    """Verifies the workspace client configuration, ensuring it has the correct product info."""

    # Using reflection to set right value for _product_info for telemetry
    product_info = getattr(ws.config, '_product_info')
    if product_info[0] != "lakebridge":
        setattr(ws.config, '_product_info', ('lakebridge', __version__))

    return ws



from pathlib import Path
import logging
from typing import Protocol

import yaml

from databricks.labs.lakebridge.connections.env_getter import EnvGetter


logger = logging.getLogger(__name__)


class SecretProvider(Protocol):
    def get_secret(self, key: str) -> str:
        pass


class LocalSecretProvider:
    def get_secret(self, key: str) -> str:
        return key


class EnvSecretProvider:
    def __init__(self, env_getter: EnvGetter):
        self._env_getter = env_getter

    def get_secret(self, key: str) -> str:
        try:
            return self._env_getter.get(str(key))
        except KeyError:
            logger.debug(f"Environment variable {key} not found. Falling back to actual value")
            return key


class DatabricksSecretProvider:
    def get_secret(self, key: str) -> str:
        raise NotImplementedError("Databricks secret vault not implemented")


class CredentialManager:
    def __init__(self, credential_loader: dict, secret_providers: dict):
        self._credentials = credential_loader
        self._secret_providers = secret_providers
        self._default_vault = self._credentials.get('secret_vault_type', 'local').lower()

    def get_credentials(self, source: str) -> dict:
        if source not in self._credentials:
            raise KeyError(f"Source system: {source} credentials not found")

        value = self._credentials[source]
        if not isinstance(value, dict):
            raise KeyError(f"Invalid credential format for source: {source}")

        return {k: self._get_secret_value(v) for k, v in value.items()}

    def _get_secret_value(self, key: str) -> str:
        provider = self._secret_providers.get(self._default_vault)
        if not provider:
            raise ValueError(f"Unsupported secret vault type: {self._default_vault}")
        return provider.get_secret(key)


def _get_home() -> Path:
    return Path(__file__).home()


def cred_file(product_name) -> Path:
    return Path(f"{_get_home()}/.databricks/labs/{product_name}/.credentials.yml")


def _load_credentials(path: Path) -> dict:
    try:
        with open(path, encoding="utf-8") as f:
            return yaml.safe_load(f)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"Credentials file not found at {path}") from e


def create_credential_manager(product_name: str, env_getter: EnvGetter):
    file_path = Path(f"{_get_home()}/.databricks/labs/{product_name}/.credentials.yml")

    secret_providers = {
        'local': LocalSecretProvider(),
        'env': EnvSecretProvider(env_getter),
        'databricks': DatabricksSecretProvider(),
    }

    loader = _load_credentials(file_path)
    return CredentialManager(loader, secret_providers)



from sqlglot.dialects.oracle import Oracle as Orc
from sqlglot.tokens import TokenType


class Oracle(Orc):
    # Instantiate Oracle Dialect
    oracle = Orc()

    class Tokenizer(Orc.Tokenizer):
        KEYWORDS = {
            **Orc.Tokenizer.KEYWORDS,
            'LONG': TokenType.TEXT,
            'NCLOB': TokenType.TEXT,
            'ROWID': TokenType.TEXT,
            'UROWID': TokenType.TEXT,
            'ANYTYPE': TokenType.TEXT,
            'ANYDATA': TokenType.TEXT,
            'ANYDATASET': TokenType.TEXT,
            'XMLTYPE': TokenType.TEXT,
            'SDO_GEOMETRY': TokenType.TEXT,
            'SDO_TOPO_GEOMETRY': TokenType.TEXT,
            'SDO_GEORASTER': TokenType.TEXT,
        }



import re
import logging
from datetime import datetime

from pyspark.errors import PySparkException
from pyspark.sql import DataFrame, DataFrameReader, SparkSession
from pyspark.sql.functions import col
from sqlglot import Dialect

from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.connectors.jdbc_reader import JDBCReaderMixin
from databricks.labs.lakebridge.reconcile.connectors.models import NormalizedIdentifier
from databricks.labs.lakebridge.reconcile.connectors.secrets import SecretsMixin
from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.recon_config import JdbcReaderOptions, Schema
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)


class OracleDataSource(DataSource, SecretsMixin, JDBCReaderMixin):
    _DRIVER = "oracle"
    _IDENTIFIER_DELIMITER = "\""
    _SCHEMA_QUERY = """select column_name, case when (data_precision is not null
                                              and data_scale <> 0)
                                              then data_type || '(' || data_precision || ',' || data_scale || ')'
                                              when (data_precision is not null and data_scale = 0)
                                              then data_type || '(' || data_precision || ')'
                                              when data_precision is null and (lower(data_type) in ('date') or
                                              lower(data_type) like 'timestamp%') then  data_type
                                              when CHAR_LENGTH = 0 then data_type
                                              else data_type || '(' || CHAR_LENGTH || ')'
                                              end data_type
                                              FROM ALL_TAB_COLUMNS
                            WHERE lower(TABLE_NAME) = '{table}' and lower(owner) = '{owner}'"""

    def __init__(
        self,
        engine: Dialect,
        spark: SparkSession,
        ws: WorkspaceClient,
        secret_scope: str,
    ):
        self._engine = engine
        self._spark = spark
        self._ws = ws
        self._secret_scope = secret_scope

    @property
    def get_jdbc_url(self) -> str:
        return (
            f"jdbc:{OracleDataSource._DRIVER}:thin:{self._get_secret('user')}"
            f"/{self._get_secret('password')}@//{self._get_secret('host')}"
            f":{self._get_secret('port')}/{self._get_secret('database')}"
        )

    def read_data(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        query: str,
        options: JdbcReaderOptions | None,
    ) -> DataFrame:
        table_query = query.replace(":tbl", f"{schema}.{table}")
        try:
            if options is None:
                return self.reader(table_query).options(**self._get_timestamp_options()).load()
            reader_options = self._get_jdbc_reader_options(options) | self._get_timestamp_options()
            df = self.reader(table_query).options(**reader_options).load()
            logger.warning(f"Fetching data using query: \n`{table_query}`")

            # Convert all column names to lower case
            df = df.select([col(c).alias(c.lower()) for c in df.columns])
            return df
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "data", table_query)

    def get_schema(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        normalize: bool = True,
    ) -> list[Schema]:
        schema_query = re.sub(
            r'\s+',
            ' ',
            OracleDataSource._SCHEMA_QUERY.format(table=table, owner=schema),
        )
        try:
            logger.debug(f"Fetching schema using query: \n`{schema_query}`")
            logger.info(f"Fetching Schema: Started at: {datetime.now()}")
            df = self.reader(schema_query).load()
            schema_metadata = df.select([col(c).alias(c.lower()) for c in df.columns]).collect()
            logger.info(f"Schema fetched successfully. Completed at: {datetime.now()}")
            logger.debug(f"schema_metadata: ${schema_metadata}")
            return [self._map_meta_column(field, normalize) for field in schema_metadata]
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "schema", schema_query)

    @staticmethod
    def _get_timestamp_options() -> dict[str, str]:
        return {
            "oracle.jdbc.mapDateToTimestamp": "False",
            "sessionInitStatement": "BEGIN dbms_session.set_nls('nls_date_format', "
            "'''YYYY-MM-DD''');dbms_session.set_nls('nls_timestamp_format', '''YYYY-MM-DD "
            "HH24:MI:SS''');END;",
        }

    def reader(self, query: str) -> DataFrameReader:
        return self._get_jdbc_reader(query, self.get_jdbc_url, OracleDataSource._DRIVER)

    def normalize_identifier(self, identifier: str) -> NormalizedIdentifier:
        return DialectUtils.normalize_identifier(
            identifier,
            source_start_delimiter=OracleDataSource._IDENTIFIER_DELIMITER,
            source_end_delimiter=OracleDataSource._IDENTIFIER_DELIMITER,
        )



import logging

logger = logging.getLogger(__name__)


class Node:
    def __init__(self, name: str):
        self.name = name.lower()
        self.children: list[str] = []
        self.parents: list[str] = []

    def add_parent(self, node: str) -> None:
        self.parents.append(node)

    def add_child(self, node: str) -> None:
        self.children.append(node)

    def __repr__(self) -> str:
        return f"Node({self.name}, {self.children})"


class DAG:
    def __init__(self):
        self.nodes: dict[str, Node] = {}

    def add_node(self, node_name: str) -> None:
        if node_name not in self.nodes and node_name not in {None, "none"}:
            self.nodes[node_name.lower()] = Node(node_name.lower())

    def add_edge(self, parent_name: str, child_name: str) -> None:
        parent_name = parent_name.lower() if parent_name is not None else None
        child_name = child_name.lower() if child_name is not None else None
        logger.debug(f"Adding edge: {parent_name} -> {child_name}")
        if parent_name not in self.nodes:
            self.add_node(parent_name)
        if child_name not in self.nodes:
            self.add_node(child_name)

        if child_name is not None:
            self.nodes[parent_name].add_child(child_name)
            self.nodes[child_name].add_parent(parent_name)

    def identify_immediate_parents(self, table_name: str) -> list[str]:
        table_name = table_name.lower()  # convert to lower() case
        if table_name not in self.nodes:
            logger.debug(f"Table with the name {table_name} not found in the DAG")
            return []

        return list(self.nodes[table_name].parents)

    def identify_immediate_children(self, table_name: str) -> list[str]:
        table_name = table_name.lower()  # convert to lower() case
        if table_name not in self.nodes:
            logger.debug(f"Table with the name {table_name} not found in the DAG")
            return []

        return list(self.nodes[table_name].children)

    def _is_root_node(self, node_name: str) -> bool:
        return len(self.identify_immediate_parents(node_name)) == 0

    def walk_bfs(self, node: Node, level: int) -> set:
        tables_at_level = set()
        queue = [(node, 0)]  # The queue for the BFS. Each element is a tuple (node, level).
        while queue:
            current_node, node_level = queue.pop(0)

            if node_level == level:
                tables_at_level.add(current_node.name)
            elif node_level > level:
                break

            for child_name in self.identify_immediate_children(current_node.name):
                queue.append((self.nodes[child_name], node_level + 1))
        return tables_at_level

    def identify_root_tables(self, level: int) -> set:
        all_nodes = set(self.nodes.values())
        root_tables_at_level = set()

        for node in all_nodes:
            if self._is_root_node(node.name):
                root_tables_at_level.update(self.walk_bfs(node, level))

        return root_tables_at_level

    def __repr__(self) -> str:
        return str({node_name: str(node) for node_name, node in self.nodes.items()})



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence, Set
from json import loads
import logging
import os
from typing import Any
from pathlib import Path

from databricks.labs.lakebridge.config import LSPConfigOptionV1

from databricks.labs.lakebridge.transpiler.lsp.lsp_engine import LSPConfig

logger = logging.getLogger(__name__)


class TranspilerRepository:
    """
    Repository for managing the installed transpilers in the user's home directory.

    The default repository for a user is always located under ~/.databricks/labs, and can be obtained
    via the `TranspilerRepository.user_home()` method.
    """

    #
    # Transpilers currently have different 'names', for historical reasons:
    #
    #  - product_name: the name of the product according to this project, assigned within the `installer` module and
    #       used as the name of the directory into which the transpiler is installed within a repository.
    #  - transpiler_name: the name of the product according to its own metadata, found in the configuration file
    #       bundled within each transpiler as distributed.
    #
    # Note: multiple installed transpilers might have the same transpiler name, but a product name is unique to a single
    # installed transpiler.
    #
    #  Known names at the moment:
    #
    #   - Morpheus:     product_name = databricks-morph-plugin,  transpiler_name = Morpheus
    #   - BladeBridge:  product_name = bladebridge,              transpiler_name = Bladebridge
    #

    @staticmethod
    def default_labs_path() -> Path:
        """Return the default path where labs applications are installed."""
        return Path.home() / ".databricks" / "labs"

    _default_repository: TranspilerRepository | None = None

    @classmethod
    def user_home(cls) -> TranspilerRepository:
        """The default repository for transpilers in the current user's home directory."""
        repository = cls._default_repository
        if repository is None:
            cls._default_repository = repository = cls(cls.default_labs_path())
        return repository

    def __init__(self, labs_path: Path) -> None:
        """Initialize the repository, based in the given location.

        This should only be used directly by tests; for the default repository, use `TranspilerRepository.user_home()`.

        Args:
            labs_path: The path where the labs applications are installed.
        """
        if self._default_repository == self and labs_path == self.default_labs_path():
            raise ValueError("Use TranspilerRepository.user_home() to get the default repository.")
        self._labs_path = labs_path

    def __repr__(self) -> str:
        return f"TranspilerRepository(labs_path={self._labs_path!r})"

    def transpilers_path(self) -> Path:
        return self._labs_path / "remorph-transpilers"

    def get_installed_version(self, product_name: str) -> str | None:
        """
        Obtain the version of an installed transpiler.

        Args:
          product_name: The product name of the transpiler whose version is sought.
        Returns:
          The version of the transpiler if it is installed, or None otherwise.
        """
        # Warning: product_name here (eg. 'morpheus') and transpiler_name elsewhere (eg. Morpheus) are not the same!
        product_path = self.transpilers_path() / product_name
        current_version_path = product_path / "state" / "version.json"
        try:
            text = current_version_path.read_text("utf-8")
        except FileNotFoundError:
            return None
        data: dict[str, Any] = loads(text)
        version: str | None = data.get("version", None)
        if not version or not version.startswith("v"):
            return None
        return version[1:]

    def all_transpiler_configs(self) -> Mapping[str, LSPConfig]:
        """Obtain all installed transpile configurations.

        Returns:
          A mapping of configurations, keyed by their transpiler names.
        """
        all_configs = self._all_transpiler_configs()
        return {config.name: config for config in all_configs}

    def all_transpiler_names(self) -> Set[str]:
        """Query the set of transpiler names for all installed transpilers."""
        all_configs = self.all_transpiler_configs()
        return frozenset(all_configs.keys())

    def all_dialects(self) -> Set[str]:
        """Query the set of dialects for all installed transpilers."""
        all_dialects: set[str] = set()
        for config in self._all_transpiler_configs():
            all_dialects = all_dialects.union(config.remorph.dialects)
        return all_dialects

    def transpilers_with_dialect(self, dialect: str) -> Set[str]:
        """
        Query the set of transpilers that can handle a given dialect.

        Args:
          dialect: The dialect to check for.
        Returns:
          The set of transpiler names for installed transpilers that can handle a given dialect.
        """
        configs = filter(lambda cfg: dialect in cfg.remorph.dialects, self.all_transpiler_configs().values())
        return frozenset(config.name for config in configs)

    def transpiler_config_path(self, transpiler_name: str) -> Path:
        """
        Obtain the path to a configuration file for an installed transpiler.

        Args:
          transpiler_name: The transpiler name of the transpiler whose path is sought.
        Returns:
          The path of the configuration file for the given installed transpiler. If multiple installed transpilers
          have the same transpiler name, either may be returned.
        Raises:
          ValueError: If there is no installed transpiler with the given transpiler name.
        """
        # Note: Because it's the transpiler name, have to hunt through the installed list rather get it directly.
        try:
            config = next(c for c in self._all_transpiler_configs() if c.name == transpiler_name)
        except StopIteration as e:
            raise ValueError(f"No such transpiler: {transpiler_name}") from e
        return config.path

    def transpiler_config_options(self, transpiler_name: str, source_dialect: str) -> Sequence[LSPConfigOptionV1]:
        """
        Query the additional configuration options that are available for a given transpiler and source dialect that it
        supports.

        Args:
          transpiler_name: The transpiler name of the transpiler that may provide options.
          source_dialect: The source dialect supported by the given for which options are sought.
        Returns:
          A sequence of configuration options, possibly empty, that are supported by the given transpiler for the
          specified source dialect.
        """
        config = self.all_transpiler_configs().get(transpiler_name, None)
        if not config:
            return []  # gracefully returns an empty list, since this can only happen during testing
        return config.options_for_dialect(source_dialect)

    def _all_transpiler_configs(self) -> Iterable[LSPConfig]:
        transpilers_path = self.transpilers_path()
        if transpilers_path.exists():
            all_files = os.listdir(transpilers_path)
            for file in all_files:
                config = self._transpiler_config(transpilers_path / file)
                if config:
                    yield config

    @classmethod
    def _transpiler_config(cls, path: Path) -> LSPConfig | None:
        if not path.is_dir() or not (path / "lib").is_dir():
            return None
        config_path = path / "lib" / "config.yml"
        if not config_path.is_file():
            return None
        try:
            return LSPConfig.load(config_path)
        except ValueError as e:
            logger.error(f"Could not load config: {path!s}", exc_info=e)
            return None



import logging
from abc import ABC, abstractmethod
from typing import Any

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine, Result, URL
from sqlalchemy.orm import sessionmaker
from sqlalchemy import text
from sqlalchemy.exc import OperationalError

logger = logging.getLogger(__name__)
logger.setLevel("INFO")


class DatabaseConnector(ABC):
    @abstractmethod
    def _connect(self) -> Engine:
        pass

    @abstractmethod
    def execute_query(self, query: str) -> Result[Any]:
        pass


class _BaseConnector(DatabaseConnector):
    def __init__(self, config: dict[str, Any]):
        self.config = config
        self.engine: Engine = self._connect()

    def _connect(self) -> Engine:
        raise NotImplementedError("Subclasses should implement this method")

    def execute_query(self, query: str) -> Result[Any]:
        if not self.engine:
            raise ConnectionError("Not connected to the database.")
        session = sessionmaker(bind=self.engine)
        connection = session()
        return connection.execute(text(query))


def _create_connector(db_type: str, config: dict[str, Any]) -> DatabaseConnector:
    connectors = {
        "snowflake": SnowflakeConnector,
        "mssql": MSSQLConnector,
        "tsql": MSSQLConnector,
        "synapse": MSSQLConnector,
    }

    connector_class = connectors.get(db_type.lower())

    if connector_class is None:
        raise ValueError(f"Unsupported database type: {db_type}")

    return connector_class(config)


class SnowflakeConnector(_BaseConnector):
    def _connect(self) -> Engine:
        raise NotImplementedError("Snowflake connector not implemented")


class MSSQLConnector(_BaseConnector):
    def _connect(self) -> Engine:
        query_params = {"driver": self.config['driver']}

        for key, value in self.config.items():
            if key not in ["user", "password", "server", "database", "port"]:
                query_params[key] = value
        connection_string = URL.create(
            "mssql+pyodbc",
            username=self.config['user'],
            password=self.config['password'],
            host=self.config['server'],
            port=self.config.get('port', 1433),
            database=self.config['database'],
            query=query_params,
        )
        return create_engine(connection_string)


class DatabaseManager:
    def __init__(self, db_type: str, config: dict[str, Any]):
        self.connector = _create_connector(db_type, config)

    def execute_query(self, query: str) -> Result[Any]:
        try:
            return self.connector.execute_query(query)
        except OperationalError:
            logger.error("Error connecting to the database check credentials")
            raise ConnectionError("Error connecting to the database check credentials") from None

    def check_connection(self) -> bool:
        query = "SELECT 101 AS test_column"
        result = self.execute_query(query)
        row = result.fetchone()
        if row is None:
            return False
        return row[0] == 101



import logging

import sqlglot.expressions as exp
from pyspark.sql import DataFrame
from sqlglot import select

from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_key_from_dialect
from databricks.labs.lakebridge.reconcile.query_builder.base import QueryBuilder
from databricks.labs.lakebridge.reconcile.query_builder.expression_generator import (
    build_column,
    build_literal,
    _get_is_string,
    build_join_clause,
)

logger = logging.getLogger(__name__)


def _union_concat(
    unions: list[exp.Select],
    result: exp.Union | exp.Select,
    cnt=0,
) -> exp.Select | exp.Union:
    if len(unions) == 1:
        return result
    if cnt == len(unions) - 2:
        return exp.union(result, unions[cnt + 1])
    cnt = cnt + 1
    res = exp.union(result, unions[cnt])
    return _union_concat(unions, res, cnt)


class SamplingQueryBuilder(QueryBuilder):
    def build_query_with_alias(self):
        self._validate(self.join_columns, "Join Columns are compulsory for sampling query")
        join_columns = self.join_columns if self.join_columns else set()

        cols = sorted((join_columns | self.select_columns) - self.threshold_columns - self.drop_columns)

        cols_with_alias = [self._build_column_with_alias(col) for col in cols]

        query = select(*cols_with_alias).from_(":tbl").where(self.filter, dialect=self.engine).sql(dialect=self.engine)

        logger.info(f"Sampling Query with Alias for {self.layer}: {query}")
        return query

    def build_query(self, df: DataFrame):
        self._validate(self.join_columns, "Join Columns are compulsory for sampling query")
        join_columns = self.join_columns if self.join_columns else set()
        if self.layer == "source":
            key_cols = sorted(join_columns)
        else:
            key_cols = sorted(self.table_conf.get_tgt_to_src_col_mapping_list(join_columns))
        keys_df = df.select(*key_cols)
        with_clause = self._get_with_clause(keys_df)

        cols = sorted((join_columns | self.select_columns) - self.threshold_columns - self.drop_columns)

        cols_with_alias = [self._build_column_with_alias(col) for col in cols]

        sql_with_transforms = self.add_transformations(cols_with_alias, self.engine)
        query_sql = select(*sql_with_transforms).from_(":tbl").where(self.filter, dialect=self.engine)
        if self.layer == "source":
            with_select = [
                build_column(this=DialectUtils.unnormalize_identifier(col), table_name="src", quoted=True)
                for col in sorted(cols)
            ]
        else:
            with_select = [
                build_column(this=DialectUtils.unnormalize_identifier(col), table_name="src", quoted=True)
                for col in sorted(self.table_conf.get_tgt_to_src_col_mapping_list(cols))
            ]

        join_clause = self._get_join_clause(key_cols)

        query = (
            with_clause.with_(alias="src", as_=query_sql)
            .select(*with_select)
            .from_("src")
            .join(join_clause)
            .sql(dialect=self.engine)
        )
        logger.info(f"Sampling Query for {self.layer}: {query}")
        return query

    def _get_join_clause(self, key_cols: list):
        normalized = [self._build_column_name_source_normalized(col) for col in key_cols]
        return build_join_clause(
            "recon", normalized, source_table_alias="src", target_table_alias="recon", kind="inner", func=exp.EQ
        )

    def _get_with_clause(self, df: DataFrame) -> exp.Select:
        union_res = []
        for row in df.collect():
            column_types = [(str(f.name).lower(), f.dataType) for f in df.schema.fields]
            column_types_dict = dict(column_types)
            orig_types_dict = {
                schema.column_name: schema.data_type
                for schema in self.schema
                if schema.column_name not in self.user_transformations
            }
            row_select = [
                (
                    build_literal(
                        this=str(value),
                        alias=DialectUtils.unnormalize_identifier(col),
                        is_string=_get_is_string(column_types_dict, col),
                        cast=orig_types_dict.get(DialectUtils.ansi_normalize_identifier(col)),
                        quoted=True,
                    )
                    if value is not None
                    else exp.Alias(this=exp.Null(), alias=DialectUtils.unnormalize_identifier(col), quoted=True)
                )
                for col, value in zip(df.columns, row)
            ]
            if get_key_from_dialect(self.engine) == "oracle":
                union_res.append(select(*row_select).from_("dual"))
            else:
                union_res.append(select(*row_select))
        union_statements = _union_concat(union_res, union_res[0], 0)
        return exp.Select().with_(alias='recon', as_=union_statements)



import logging
import typing as t
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

from sqlglot import expressions as exp, parse, transpile, Dialect
from sqlglot.errors import ErrorLevel, ParseError, TokenError, UnsupportedError
from sqlglot.expressions import Expression
from sqlglot.tokens import Token, TokenType

from databricks.labs.lakebridge.config import TranspileResult, TranspileConfig
from databricks.labs.lakebridge.helpers.file_utils import is_sql_file
from databricks.labs.lakebridge.helpers.string_utils import format_error_message
from databricks.labs.lakebridge.transpiler.sqlglot import lca_utils
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import get_dialect
from databricks.labs.lakebridge.transpiler.sqlglot.dialect_utils import SQLGLOT_DIALECTS
from databricks.labs.lakebridge.transpiler.transpile_status import TranspileError, ErrorKind, ErrorSeverity
from databricks.labs.lakebridge.transpiler.transpile_engine import TranspileEngine

logger = logging.getLogger(__name__)


@dataclass
class ParsedExpression:
    original_sql: str
    parsed_expression: Expression


@dataclass
class ParserProblem:
    original_sql: str
    transpile_error: TranspileError


class SqlglotEngine(TranspileEngine):

    @property
    def supported_dialects(self) -> list[str]:
        return sorted(SQLGLOT_DIALECTS.keys())

    @property
    def transpiler_name(self) -> str:
        return "sqlglot"

    def _partial_transpile(
        self,
        read_dialect: Dialect,
        write_dialect: Dialect,
        source_code: str,
        file_path: Path,
    ) -> tuple[list[str], list[ParserProblem]]:
        transpiled_sqls: list[str] = []
        parsed_expressions, problem_list = self.safe_parse(read_dialect, source_code, file_path)
        for parsed_expression in parsed_expressions:
            try:
                transpiled_sql = write_dialect.generate(parsed_expression.parsed_expression, pretty=True)
                # Checking if the transpiled SQL is a comment and raise an error
                if transpiled_sql.startswith("--"):
                    raise UnsupportedError("Unsupported SQL")
                transpiled_sqls.append(transpiled_sql)
            except TokenError as e:
                error_msg = format_error_message("Token Error", e, parsed_expression.original_sql)
                error = TranspileError("TOKEN_ERROR", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
                problem_list.append(ParserProblem(parsed_expression.original_sql, error))
            except ParseError as e:
                error_msg = format_error_message("Parsing Error", e, parsed_expression.original_sql)
                error = TranspileError("PARSE_ERROR", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
                problem_list.append(ParserProblem(parsed_expression.original_sql, error))
            except UnsupportedError as e:
                error_msg = format_error_message("Unsupported SQL Error", e, parsed_expression.original_sql)
                error = TranspileError("UNSUPPORTED_SQL", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
                problem_list.append(ParserProblem(parsed_expression.original_sql, error))
        return transpiled_sqls, problem_list

    async def initialize(self, config: TranspileConfig) -> None:
        pass

    async def shutdown(self) -> None:
        pass

    async def transpile(
        self, source_dialect: str, target_dialect: str, source_code: str, file_path: Path
    ) -> TranspileResult:
        read_dialect = get_dialect(source_dialect)
        error: TranspileError | None = self._check_supported(read_dialect, source_code, file_path)
        if error:
            return TranspileResult(source_code, 1, [error])
        write_dialect = get_dialect(target_dialect)
        try:
            transpiled_expressions = transpile(
                source_code, read=read_dialect, write=write_dialect, pretty=True, error_level=ErrorLevel.RAISE
            )
            transpiled_code = "\n".join(transpiled_expressions)
            return TranspileResult(transpiled_code, len(transpiled_expressions), [])
        except (ParseError, TokenError, UnsupportedError) as e:
            logger.error(f"Exception caught for file {file_path!s}: {e}")
            transpiled_expressions, problems = self._partial_transpile(
                read_dialect, write_dialect, source_code, file_path
            )
            transpiled_code = "\n".join(transpiled_expressions)
            return TranspileResult(transpiled_code, 1, [problem.transpile_error for problem in problems])

    def parse(
        self, source_dialect: str, source_sql: str, file_path: Path
    ) -> tuple[list[Expression | None] | None, TranspileError | None]:
        expression = None
        error = None
        try:
            expression = parse(source_sql, read=source_dialect, error_level=ErrorLevel.IMMEDIATE)
        except TokenError as e:
            error_msg = format_error_message("Token Error", e, source_sql)
            error = TranspileError("TOKEN_ERROR", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
        except ParseError as e:
            error_msg = format_error_message("Parsing Error", e, source_sql)
            error = TranspileError("PARSE_ERROR", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
        except UnsupportedError as e:
            error_msg = format_error_message("Unsupported SQL Error", e, source_sql)
            error = TranspileError("UNSUPPORTED_SQL", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
        return expression, error

    def analyse_table_lineage(
        self, source_dialect: str, source_code: str, file_path: Path
    ) -> Iterable[tuple[str, str]]:
        parsed_expression, _ = self.parse(source_dialect, source_code, file_path)
        if parsed_expression is not None:
            for expr in parsed_expression:
                child: str = str(file_path)
                if expr is not None:
                    # TODO: fix possible issue where the file reference is lost (if we have a 'create')
                    for change in expr.find_all(exp.Create, exp.Insert, exp.Merge, bfs=False):
                        child = self._find_root_table(change)

                    for query in expr.find_all(exp.Select, exp.Join, exp.With, bfs=False):
                        table = self._find_root_table(query)
                        if table:
                            yield table, child

    def safe_parse(
        self, read_dialect: Dialect, source_code: str, file_path: Path
    ) -> tuple[list[ParsedExpression], list[ParserProblem]]:
        try:
            tokens = read_dialect.tokenize(sql=source_code)
            return self._safe_parse(read_dialect, tokens, file_path)
        except TokenError as e:
            error_msg = format_error_message("Token error", e, source_code)
            error = TranspileError("TOKEN_ERROR", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
            return [], [ParserProblem(source_code, error)]

    def _safe_parse(
        self, read_dialect: Dialect, all_tokens: list[Token], file_path: Path
    ) -> tuple[list[ParsedExpression], list[ParserProblem]]:
        chunks = self._make_chunks(all_tokens)
        parsed_expressions: list[ParsedExpression] = []
        problems: list[ParserProblem] = []
        parser_opts = {"error_level": ErrorLevel.RAISE}
        parser = read_dialect.parser(**parser_opts)
        for sql, tokens in chunks:
            try:
                expressions = parser.parse(tokens)
                expression = t.cast(Expression, expressions[0])
                parsed_expressions.append(ParsedExpression(sql, expression))
            except TokenError as e:
                error_msg = format_error_message("Token error", e, sql)
                error = TranspileError("TOKEN_ERROR", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
                problems.append(ParserProblem(sql, error))
            except ParseError as e:
                error_msg = format_error_message("Parsing error", e, sql)
                error = TranspileError("PARSE_ERROR", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
                problems.append(ParserProblem(sql, error))
            except UnsupportedError as e:
                error_msg = format_error_message("Unsupported SQL error", e, sql)
                error = TranspileError("UNSUPPORTED_SQL", ErrorKind.PARSING, ErrorSeverity.ERROR, file_path, error_msg)
                problems.append(ParserProblem(sql, error))
            finally:
                parser.reset()
        return parsed_expressions, problems

    @staticmethod
    def _make_chunks(tokens: list[Token]) -> list[tuple[str, list[Token]]]:
        chunks: list[tuple[str, list[Token]]] = []
        current_chunk: list[Token] = []
        # Split tokens into chunks based on semicolons(or other separators)
        # Need to define the separator in Class Tokenizer
        for token in tokens:
            current_chunk.append(token)
            if token.token_type in {TokenType.SEMICOLON}:
                original_sql = " ".join([token.text for token in current_chunk]).strip()
                chunks.append((original_sql, current_chunk))
                # reset
                current_chunk = []
        # don't forget the last chunk
        if current_chunk:
            original_sql = " ".join([token.text for token in current_chunk]).strip()
            chunks.append((original_sql, current_chunk))
        return chunks

    @staticmethod
    def _find_root_table(expression) -> str:
        table = expression.find(exp.Table, bfs=False)
        return table.name if table else ""

    def _check_supported(self, source_dialect: Dialect, source_code: str, file_path: Path) -> TranspileError | None:
        return lca_utils.check_for_unsupported_lca(source_dialect, source_code, file_path)

    def is_supported_file(self, file: Path) -> bool:
        return is_sql_file(file)



import os
from dataclasses import dataclass
from duckdb import DuckDBPyConnection

from databricks.labs.lakebridge.assessments.pipeline import PipelineClass

PROFILER_DB_NAME = "profiler_extract.db"


@dataclass(frozen=True)
class ValidationOutcome:
    """A data class that holds the outcome of a table validation check."""

    table: str
    column: str | None
    strategy: str
    outcome: str
    severity: str


class ValidationStrategy:
    """Abstract class for validating a Profiler table"""

    def validate(self, connection: DuckDBPyConnection) -> ValidationOutcome:
        raise NotImplementedError


class NullValidationCheck(ValidationStrategy):
    """Concrete class for validating null values in a profiler table"""

    def __init__(self, table, column, severity="WARN"):
        self.name = self.__class__.__name__
        self.table = table
        self.column = column
        self.severity = severity

    def validate(self, connection: DuckDBPyConnection) -> ValidationOutcome:
        """
        Validates that a column does not contain null values.
        input:
          connection: a DuckDB connection object
        """
        result = connection.execute(f"SELECT COUNT(*) FROM {self.table} WHERE {self.column} IS NULL").fetchone()
        if result:
            row_count = result[0]
            outcome = "FAIL" if row_count > 0 else "PASS"
        else:
            outcome = "FAIL"
        return ValidationOutcome(self.table, self.column, self.name, outcome, self.severity)


class EmptyTableValidationCheck(ValidationStrategy):
    """Concrete class for validating empty tables from a profiler run."""

    def __init__(self, table, severity="WARN"):
        self.name = self.__class__.__name__
        self.table = table
        self.severity = severity

    def validate(self, connection) -> ValidationOutcome:
        """Validates that a table is not empty.
        input:
          connection: a DuckDB connection object
        returns:
          a ValidationOutcome object
        """
        result = connection.execute(f"SELECT COUNT(*) FROM {self.table}").fetchone()
        if result:
            row_count = result[0]
            outcome = "PASS" if row_count > 0 else "FAIL"
        else:
            outcome = "FAIL"
        return ValidationOutcome(self.table, None, self.name, outcome, self.severity)


def get_profiler_extract_path(pipeline_config_path: str) -> str:
    """
    Returns the filesystem path of the profiler extract database.
    input:
       pipeline_config_path: the location of the pipeline definition .yml file
    returns:
       the filesystem path to the profiler extract database
    """
    pipeline_config = PipelineClass.load_config_from_yaml(pipeline_config_path)
    normalized_db_path = os.path.normpath(pipeline_config.extract_folder)
    database_path = f"{normalized_db_path}/{PROFILER_DB_NAME}"
    return database_path


def build_validation_report(
    validations: list[ValidationStrategy], connection: DuckDBPyConnection
) -> list[ValidationOutcome]:
    """
    Builds a list of ValidationOutcomes from list of validation checks.
    input:
      validations: a list of ValidationStrategy objects
      connection: a DuckDB connection object
    returns: a list of ValidationOutcomes
    """
    validation_report = []
    for validation in validations:
        validation_report.append(validation.validate(connection))
    return validation_report



import inspect
import logging
import time
from functools import wraps

logger = logging.getLogger(__name__)


def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        name = inspect.getmodule(func).__name__.split(".")[3].capitalize()
        logger.info(f"{name} took {total_time:.4f} seconds")
        return result

    return timeit_wrapper



from pathlib import Path

from databricks.labs.lakebridge.coverage import commons

if __name__ == "__main__":
    output_dir = commons.get_env_var("OUTPUT_DIR", required=True)
    if not output_dir:
        raise ValueError("Environment variable `OUTPUT_DIR` is required")
    commons.local_report(Path(output_dir))



import logging

from databricks.labs.blueprint.entrypoint import is_in_debug
from databricks.sdk import WorkspaceClient

from databricks.labs.lakebridge.__about__ import __version__
from databricks.labs.lakebridge.contexts.application import ApplicationContext

logger = logging.getLogger("databricks.labs.lakebridge.install")


def run(context: ApplicationContext):
    context.workspace_installation.uninstall(context.remorph_config)


if __name__ == "__main__":
    logger.setLevel("INFO")
    if is_in_debug():
        logging.getLogger("databricks").setLevel(logging.DEBUG)

    run(
        ApplicationContext(
            WorkspaceClient(
                product="lakebridge",
                product_version=__version__,
            )
        )
    )



from dataclasses import dataclass, field
from typing import Any


@dataclass
class TableFQN:
    catalog: str | None
    schema: str
    name: str

    @property
    def fqn(self) -> str:
        if self.catalog:
            return f"{self.catalog}.{self.schema}.{self.name}"
        return f"{self.schema}.{self.name}"


@dataclass
class FieldInfo:
    name: str
    data_type: str
    nullable: bool | None = None
    metadata: dict[str, Any] | None = None
    comment: str | None = None


@dataclass
class TableDefinition:
    fqn: TableFQN
    location: str | None = None
    table_format: str | None = None
    view_text: str | None = None
    columns: list[FieldInfo] = field(default_factory=list)
    primary_keys: list[str] | None = None
    size_gb: int | None = None
    comment: str | None = None



import logging
import re

from sqlglot import expressions as exp
from sqlglot.dialects.dialect import build_date_delta as parse_date_delta, build_formatted_time
from sqlglot.dialects.snowflake import Snowflake as SqlglotSnowflake
from sqlglot.errors import TokenError, ParseError
from sqlglot.helper import is_int, seq_get
from sqlglot.optimizer.simplify import simplify_literals
from sqlglot.parser import build_var_map as parse_var_map
from sqlglot.tokens import Token, TokenType
from sqlglot.trie import new_trie

from databricks.labs.lakebridge.transpiler.sqlglot import local_expression

logger = logging.getLogger(__name__)
# pylint: disable=protected-access
""" SF Supported Date and Time Parts:
    https://docs.snowflake.com/en/sql-reference/functions-date-time#label-supported-date-time-parts
    Covers DATEADD, DATEDIFF, DATE_TRUNC, LAST_DAY
"""
DATE_DELTA_INTERVAL = {
    "years": "year",
    "year": "year",
    "yrs": "year",
    "yr": "year",
    "yyyy": "year",
    "yyy": "year",
    "yy": "year",
    "y": "year",
    "quarters": "quarter",
    "quarter": "quarter",
    "qtrs": "quarter",
    "qtr": "quarter",
    "q": "quarter",
    "months": "month",
    "month": "month",
    "mons": "month",
    "mon": "month",
    "mm": "month",
    "weekofyear": "week",
    "week": "week",
    "woy": "week",
    "wy": "week",
    "wk": "week",
    "w": "week",
    "dayofmonth": "day",
    "days": "day",
    "day": "day",
    "dd": "day",
    "d": "day",
}

rank_functions = (
    local_expression.CumeDist,
    exp.FirstValue,
    exp.LastValue,
    local_expression.NthValue,
    local_expression.Ntile,
)


def _parse_to_timestamp(args: list) -> exp.StrToTime | exp.UnixToTime | exp.TimeStrToTime:
    if len(args) == 2:
        first_arg, second_arg = args
        if second_arg.is_string:
            # case: <string_expr> [ , <format> ]
            return build_formatted_time(exp.StrToTime, "snowflake", default=True)(args)
        return exp.UnixToTime(this=first_arg, scale=second_arg)

    # The first argument might be an expression like 40 * 365 * 86400, so we try to
    # reduce it using `simplify_literals` first and then check if it's a Literal.
    first_arg = seq_get(args, 0)
    if not isinstance(simplify_literals(first_arg, root=True), exp.Literal):
        # case: <variant_expr> or other expressions such as columns
        return exp.TimeStrToTime.from_arg_list(args)

    if first_arg.is_string:
        if is_int(first_arg.this):
            # case: <integer>
            return exp.UnixToTime.from_arg_list(args)

        # case: <date_expr>
        return build_formatted_time(exp.StrToTime, "snowflake", default=True)(args)

    # case: <numeric_expr>
    return exp.UnixToTime.from_arg_list(args)


def _parse_date_add(args: list) -> exp.DateAdd:
    return exp.DateAdd(this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0))


def _parse_split_part(args: list) -> local_expression.SplitPart:
    if len(args) != 3:
        err_msg = f"Error Parsing args `{args}`. Number of args must be 3, given {len(args)}"
        raise ParseError(err_msg)
    part_num_literal = seq_get(args, 2)
    part_num_if = None
    if isinstance(part_num_literal, exp.Literal):
        # In Snowflake if the partNumber is 0, it is treated as 1.
        # Please refer to https://docs.snowflake.com/en/sql-reference/functions/split_part
        if part_num_literal.is_int and int(part_num_literal.name) == 0:
            part_num_literal = exp.Literal.number(1)
    else:
        cond = exp.EQ(this=part_num_literal, expression=exp.Literal.number(0))
        part_num_if = exp.If(this=cond, true=exp.Literal.number(1), false=part_num_literal)

    part_num = part_num_if if part_num_if is not None else part_num_literal
    return local_expression.SplitPart(this=seq_get(args, 0), expression=seq_get(args, 1), partNum=part_num)


def _div0_to_if(args: list) -> exp.If:
    cond = exp.EQ(this=seq_get(args, 1), expression=exp.Literal.number(0))
    true = exp.Literal.number(0)
    false = exp.Div(this=seq_get(args, 0), expression=seq_get(args, 1))
    return exp.If(this=cond, true=true, false=false)


def _div0null_to_if(args: list) -> exp.If:
    cond = exp.Or(
        this=exp.EQ(this=seq_get(args, 1), expression=exp.Literal.number(0)),
        expression=exp.Is(this=seq_get(args, 1), expression=exp.Null()),
    )
    true = exp.Literal.number(0)
    false = exp.Div(this=seq_get(args, 0), expression=seq_get(args, 1))
    return exp.If(this=cond, true=true, false=false)


def _parse_json_extract_path_text(args: list) -> local_expression.JsonExtractPathText:
    if len(args) != 2:
        err_message = f"Error Parsing args `{args}`. Number of args must be 2, given {len(args)}"
        raise ParseError(err_message)
    return local_expression.JsonExtractPathText(this=seq_get(args, 0), path_name=seq_get(args, 1))


def _parse_array_contains(args: list) -> exp.ArrayContains:
    if len(args) != 2:
        err_message = f"Error Parsing args `{args}`. Number of args must be 2, given {len(args)}"
        raise ParseError(err_message)
    return exp.ArrayContains(this=seq_get(args, 1), expression=seq_get(args, 0))


def _parse_dayname(args: list) -> local_expression.DateFormat:
    """
        * E, EE, EEE, returns short day name (Mon)
        * EEEE, returns full day name (Monday)
    :param args: node expression
    :return: DateFormat with `E` format
    """
    if len(args) != 1:
        err_message = f"Error Parsing args `{args}`. Number of args must be 1, given {len(args)}"
        raise ParseError(err_message)
    return local_expression.DateFormat(this=seq_get(args, 0), expression=exp.Literal.string("E"))


def _parse_trytonumber(args: list) -> local_expression.TryToNumber:
    if len(args) == 1:
        msg = f"""*Warning:: Parsing args `{args}`:
                             * `format` is missing
                             * assuming defaults `precision`[38] and `scale`[0]
                          """
        logger.warning(msg)
    elif len(args) == 3:
        msg = f"""Error Parsing args `{args}`:
                             * `format` is required
                             * `precision` and `scale` both are required [if specified]
                          """
        raise ParseError(msg)

    if len(args) == 4:
        return local_expression.TryToNumber(
            this=seq_get(args, 0), expression=seq_get(args, 1), precision=seq_get(args, 2), scale=seq_get(args, 3)
        )

    return local_expression.TryToNumber(this=seq_get(args, 0), expression=seq_get(args, 1))


def _parse_monthname(args: list) -> local_expression.DateFormat:
    if len(args) != 1:
        err_message = f"Error Parsing args `{args}`. Number of args must be 1, given {len(args)}"
        raise ParseError(err_message)
    return local_expression.DateFormat(this=seq_get(args, 0), expression=exp.Literal.string("MMM"))


def _parse_object_construct(args: list) -> exp.StarMap | exp.Struct:
    expression = parse_var_map(args)

    if isinstance(expression, exp.StarMap):
        return exp.Struct(expressions=[expression.this])

    return exp.Struct(
        expressions=[
            exp.PropertyEQ(this=k.this, expression=v) for k, v in zip(expression.keys, expression.values, strict=False)
        ]
    )


def _parse_to_boolean(args: list, *, error=False) -> local_expression.ToBoolean:
    this_arg = seq_get(args, 0)
    return local_expression.ToBoolean(this=this_arg, raise_error=exp.Literal.number(1 if error else 0))


def _parse_tonumber(args: list) -> local_expression.ToNumber:
    if len(args) > 4:
        error_msg = f"""Error Parsing args args:
                          * Number of args cannot be more than `4`, given `{len(args)}`
                          """
        raise ParseError(error_msg)

    match len(args):
        case 1:
            msg = (
                "Precision and Scale are not specified, assuming defaults `precision`[38] and `scale`[0]. "
                "If Format is not specified, it will be inferred as simple cast as decimal"
            )
            logger.warning(msg)
            return local_expression.ToNumber(this=seq_get(args, 0))
        case 3:
            msg = "If Format is not specified, it will be inferred as simple cast as decimal"
            logger.warning(msg)
            return local_expression.ToNumber(this=seq_get(args, 0), precision=seq_get(args, 1), scale=seq_get(args, 2))
        case 4:
            return local_expression.ToNumber(
                this=seq_get(args, 0), expression=seq_get(args, 1), precision=seq_get(args, 2), scale=seq_get(args, 3)
            )

    return local_expression.ToNumber(this=seq_get(args, 0), expression=seq_get(args, 1))


def contains_expression(expr, target_type):
    if isinstance(expr, target_type):
        return True
    if hasattr(expr, 'this') and contains_expression(expr.this, target_type):
        return True
    if hasattr(expr, 'expressions'):
        for sub_expr in expr.expressions:
            if contains_expression(sub_expr, target_type):
                return True
    return False


def _parse_sha2(args: list) -> exp.SHA2:
    if len(args) == 1:
        return exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(256))
    return exp.SHA2(this=seq_get(args, 0), length=seq_get(args, 1))


def _parse_last_day(args: list) -> exp.LastDay | exp.DateSub:
    if len(args) == 1:
        return exp.LastDay.from_arg_list(args)

    date_part = seq_get(args, 1)
    if isinstance(date_part, exp.Literal):
        date_part = DATE_DELTA_INTERVAL.get(date_part.this.lower(), None)
    elif isinstance(date_part, exp.Column):
        date_part = DATE_DELTA_INTERVAL.get(date_part.name.lower(), None)

    if date_part is None or date_part.lower() not in ('year', 'quarter', 'month', 'week'):
        raise ParseError(f'Invalid date part {date_part} for last_day')

    date_trunc_expr = local_expression.DateTrunc(this=seq_get(args, 0), unit=exp.Literal.string(date_part))
    # Add one date part
    date_add_expr = parse_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL)(
        [exp.Literal.string(date_part), exp.Literal.number(1), date_trunc_expr]
    )

    # Subtract one day
    date_sub_expr = parse_date_delta(exp.DateSub, unit_mapping=DATE_DELTA_INTERVAL)(
        [exp.Literal.string('DAY'), exp.Literal.number(1), date_add_expr]
    )

    return date_sub_expr


class Snowflake(SqlglotSnowflake):
    # Instantiate Snowflake Dialect
    snowflake = SqlglotSnowflake()

    class Tokenizer(SqlglotSnowflake.Tokenizer):

        COMMENTS = ["--", "//", ("/*", "*/")]
        STRING_ESCAPES = ["\\", "'"]

        CUSTOM_TOKEN_MAP = {
            r"(?i)CREATE\s+OR\s+REPLACE\s+PROCEDURE": TokenType.PROCEDURE,
            r"(?i)var\s+\w+\s+=\s+\w+?": TokenType.VAR,
        }

        SINGLE_TOKENS = {
            **SqlglotSnowflake.Tokenizer.SINGLE_TOKENS,
            "&": TokenType.PARAMETER,  # https://docs.snowflake.com/en/user-guide/snowsql-use#substituting-variables-in-a-session
            "!": TokenType.COMMAND,
        }

        KEYWORDS = {**SqlglotSnowflake.Tokenizer.KEYWORDS}
        # DEC is not a reserved keyword in Snowflake it can be used as table alias
        KEYWORDS.pop("DEC")

        @classmethod
        def update_keywords(cls, new_key_word_dict):
            cls.KEYWORDS = new_key_word_dict | cls.KEYWORDS

        @classmethod
        def merge_trie(cls, parent_trie, curr_trie):
            merged_trie = {}
            logger.debug(f"The Parent Trie is {parent_trie}")
            logger.debug(f"The Input Trie is {curr_trie}")
            for key in set(parent_trie.keys()) | set(curr_trie.keys()):  # Get all unique keys from both tries
                if key in parent_trie and key in curr_trie:  # If the key is in both tries, merge the subtries
                    if isinstance(parent_trie[key], dict) and isinstance(curr_trie[key], dict):
                        logger.debug(f"New trie inside the key is {curr_trie}")
                        logger.debug(f"Parent trie inside the key is {parent_trie}")
                        merged_trie[key] = cls.merge_trie(parent_trie[key], curr_trie[key])
                        logger.debug(f"Merged Trie is {merged_trie}")
                    elif isinstance(parent_trie[key], dict):
                        merged_trie[key] = parent_trie[key]
                    else:
                        merged_trie[key] = curr_trie[key]
                elif key in parent_trie:  # If the key is only in trie1, add it to the merged trie
                    merged_trie[key] = parent_trie[key]
                else:  # If the key is only in trie2, add it to the merged trie
                    merged_trie[key] = curr_trie[key]
            return merged_trie

        @classmethod
        def update_keyword_trie(
            cls,
            curr_trie,
            parent_trie=None,
        ):
            if parent_trie is None:
                parent_trie = cls._KEYWORD_TRIE
            cls.KEYWORD_TRIE = cls.merge_trie(parent_trie, curr_trie)

        def match_strings_token_dict(self, string, pattern_dict):
            result_dict = {}
            for pattern in pattern_dict:
                matches = re.finditer(pattern, string, re.MULTILINE | re.IGNORECASE | re.DOTALL)
                for _, match in enumerate(matches, start=1):
                    result_dict[match.group().upper()] = pattern_dict[pattern]
            return result_dict

        def match_strings_list(self, string, pattern_dict):
            result = []
            for pattern in pattern_dict:
                matches = re.finditer(pattern, string, re.MULTILINE | re.IGNORECASE | re.DOTALL)
                for _, match in enumerate(matches, start=1):
                    result.append(match.group().upper())
            return result

        def tokenize(self, sql: str) -> list[Token]:
            """Returns a list of tokens corresponding to the SQL string `sql`."""
            self.reset()
            self.sql = sql
            # Update Keywords
            ref_dict = self.match_strings_token_dict(sql, self.CUSTOM_TOKEN_MAP)
            self.update_keywords(ref_dict)
            # Update Keyword Trie
            custom_trie = new_trie(self.match_strings_list(sql, self.CUSTOM_TOKEN_MAP))
            logger.debug(
                f"The New Trie after adding the REF, VAR and IF ELSE blocks "
                f"based on {self.CUSTOM_TOKEN_MAP}, is \n\n {custom_trie}"
            )
            self.update_keyword_trie(custom_trie)
            logger.debug(f"Updated New Trie is {custom_trie}")
            # Parent Code
            self.size = len(sql)
            try:
                self._scan()
            except Exception as e:
                start = self._current - 50
                end = self._current + 50
                start = start if start > 0 else 0
                end = end if end < self.size else self.size - 1
                context = self.sql[start:end]
                msg = f"Error tokenizing '{context}'"
                raise TokenError(msg) from e
            return self.tokens

    class Parser(SqlglotSnowflake.Parser):
        FUNCTIONS = {
            **SqlglotSnowflake.Parser.FUNCTIONS,
            "ARRAY_AGG": exp.ArrayAgg.from_arg_list,
            "STRTOK_TO_ARRAY": local_expression.Split.from_arg_list,
            "DATE_FROM_PARTS": local_expression.MakeDate.from_arg_list,
            "CONVERT_TIMEZONE": local_expression.ConvertTimeZone.from_arg_list,
            "TRY_TO_DATE": local_expression.TryToDate.from_arg_list,
            "TRY_TO_TIMESTAMP": local_expression.TryToTimestamp.from_arg_list,
            "STRTOK": local_expression.StrTok.from_arg_list,
            "SPLIT_PART": _parse_split_part,
            "TIMESTAMPADD": _parse_date_add,
            "TRY_TO_DECIMAL": _parse_trytonumber,
            "TRY_TO_NUMBER": _parse_trytonumber,
            "TRY_TO_NUMERIC": _parse_trytonumber,
            "DATEADD": parse_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
            "DATEDIFF": parse_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
            "IS_INTEGER": local_expression.IsInteger.from_arg_list,
            "DIV0": _div0_to_if,
            "DIV0NULL": _div0null_to_if,
            "JSON_EXTRACT_PATH_TEXT": _parse_json_extract_path_text,
            "BITOR_AGG": local_expression.BitOr.from_arg_list,
            "ARRAY_CONTAINS": _parse_array_contains,
            "DAYNAME": _parse_dayname,
            "BASE64_ENCODE": exp.ToBase64.from_arg_list,
            "BASE64_DECODE_STRING": exp.FromBase64.from_arg_list,
            "TRY_BASE64_DECODE_STRING": exp.FromBase64.from_arg_list,
            "ARRAY_CONSTRUCT_COMPACT": local_expression.ArrayConstructCompact.from_arg_list,
            "ARRAY_INTERSECTION": local_expression.ArrayIntersection.from_arg_list,
            "ARRAY_SLICE": local_expression.ArraySlice.from_arg_list,
            "MONTHNAME": _parse_monthname,
            "MONTH_NAME": _parse_monthname,
            "OBJECT_CONSTRUCT": _parse_object_construct,
            "OBJECT_KEYS": local_expression.ObjectKeys.from_arg_list,
            "TRY_PARSE_JSON": exp.ParseJSON.from_arg_list,
            "TIMEDIFF": parse_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
            "TIMESTAMPDIFF": parse_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
            "TIMEADD": _parse_date_add,
            "TO_BOOLEAN": lambda args: _parse_to_boolean(args, error=True),
            "TO_DECIMAL": _parse_tonumber,
            "TO_DOUBLE": local_expression.ToDouble.from_arg_list,
            "TO_NUMBER": _parse_tonumber,
            "TO_NUMERIC": _parse_tonumber,
            "TO_OBJECT": local_expression.ToObject.from_arg_list,
            "TO_TIME": _parse_to_timestamp,
            "TIMESTAMP_FROM_PARTS": local_expression.TimestampFromParts.from_arg_list,
            "TO_VARIANT": local_expression.ToVariant.from_arg_list,
            "TRY_TO_BOOLEAN": lambda args: _parse_to_boolean(args, error=False),
            "UUID_STRING": local_expression.UUID.from_arg_list,
            "SYSDATE": exp.CurrentTimestamp.from_arg_list,
            "TRUNC": lambda args: local_expression.DateTrunc(unit=seq_get(args, 1), this=seq_get(args, 0)),
            "APPROX_PERCENTILE": exp.ApproxQuantile.from_arg_list,
            "NTH_VALUE": local_expression.NthValue.from_arg_list,
            "MEDIAN": local_expression.Median.from_arg_list,
            "CUME_DIST": local_expression.CumeDist.from_arg_list,
            "DENSE_RANK": local_expression.DenseRank.from_arg_list,
            "RANK": local_expression.Rank.from_arg_list,
            "PERCENT_RANK": local_expression.PercentRank.from_arg_list,
            "NTILE": local_expression.Ntile.from_arg_list,
            "TO_ARRAY": local_expression.ToArray.from_arg_list,
            "SHA2": _parse_sha2,
            "LAST_DAY": _parse_last_day,
            "ARRAY_FLATTEN": exp.Flatten.from_arg_list,
        }

        FUNCTION_PARSERS = {
            **SqlglotSnowflake.Parser.FUNCTION_PARSERS,
            "LISTAGG": lambda self: self._parse_list_agg(),
        }

        PLACEHOLDER_PARSERS = {
            **SqlglotSnowflake.Parser.PLACEHOLDER_PARSERS,
            TokenType.PARAMETER: lambda self: self._parse_parameter(),
        }

        FUNC_TOKENS = {*SqlglotSnowflake.Parser.FUNC_TOKENS, TokenType.COLLATE}

        COLUMN_OPERATORS = {
            **SqlglotSnowflake.Parser.COLUMN_OPERATORS,
        }

        TIMESTAMPS: set[TokenType] = SqlglotSnowflake.Parser.TIMESTAMPS.copy() - {TokenType.TIME}

        RANGE_PARSERS = {
            **SqlglotSnowflake.Parser.RANGE_PARSERS,
        }

        ALTER_PARSERS = {**SqlglotSnowflake.Parser.ALTER_PARSERS}

        def _parse_list_agg(self) -> exp.GroupConcat:
            if self._match(TokenType.DISTINCT):
                args: list[exp.Expression] = [self.expression(exp.Distinct, expressions=[self._parse_conjunction()])]
                if self._match(TokenType.COMMA):
                    args.extend(self._parse_csv(self._parse_conjunction))
            else:
                args = self._parse_csv(self._parse_conjunction)

            return self.expression(exp.GroupConcat, this=args[0], separator=seq_get(args, 1))

        def _parse_types(
            self, check_func: bool = False, schema: bool = False, allow_identifiers: bool = True
        ) -> exp.Expression | None:
            this = super()._parse_types(check_func=check_func, schema=schema, allow_identifiers=allow_identifiers)
            # https://docs.snowflake.com/en/sql-reference/data-types-numeric Numeric datatype alias
            if (
                isinstance(this, exp.DataType)
                and this.is_type("numeric", "decimal", "number", "integer", "int", "smallint", "bigint")
                and not this.expressions
            ):
                return exp.DataType.build("DECIMAL(38,0)")
            return this

        def _parse_parameter(self):
            wrapped = self._match(TokenType.L_BRACE)
            this = self._parse_var() or self._parse_identifier() or self._parse_primary()
            self._match(TokenType.R_BRACE)
            suffix: exp.Expression | None = None
            if not self._match(TokenType.SPACE) or self._match(TokenType.DOT):
                suffix = self._parse_var() or self._parse_identifier() or self._parse_primary()

            return self.expression(local_expression.Parameter, this=this, wrapped=wrapped, suffix=suffix)

        def _parse_window(self, this: exp.Expression | None, alias: bool = False) -> exp.Expression | None:
            window = super()._parse_window(this=this, alias=alias)
            # Adding default window frame for the rank-related functions in snowflake
            if window and contains_expression(window.this, rank_functions) and window.args.get('spec') is None:
                window.args['spec'] = self.expression(
                    exp.WindowSpec,
                    kind="ROWS",
                    start="UNBOUNDED",
                    start_side="PRECEDING",
                    end="UNBOUNDED",
                    end_side="FOLLOWING",
                )
            return window

        def _parse_alter_table_add(self) -> list[exp.Expression]:
            index = self._index - 1
            if self._match_set(self.ADD_CONSTRAINT_TOKENS, advance=False):
                return self._parse_csv(
                    lambda: self.expression(exp.AddConstraint, expressions=self._parse_csv(self._parse_constraint))
                )

            self._retreat(index)
            if not self.ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN and self._match_text_seq("ADD"):
                return self._parse_wrapped_csv(self._parse_field_def, optional=True)

            if self._match_text_seq("ADD", "COLUMN"):
                schema = self._parse_schema()
                if schema:
                    return [schema]
                # return self._parse_csv in case of COLUMNS are not enclosed in brackets ()
                return self._parse_csv(self._parse_field_def)

            return self._parse_wrapped_csv(self._parse_add_column, optional=True)



import logging
from abc import ABC, abstractmethod

from pyspark.sql import DataFrame
from pyspark.sql import functions as F

from databricks.labs.lakebridge.reconcile.constants import SamplingOptionMethod, SamplingSpecificationsType
from databricks.labs.lakebridge.reconcile.recon_config import SamplingOptions, SamplingSpecifications

logger = logging.getLogger(__name__)

_MIN_SAMPLE_COUNT = 50
_MAX_SAMPLE_COUNT = 400

_MIN_BUCKET_LIMIT = 2
_MAX_BUCKET_LIMIT = 50


class Sampler(ABC):
    def __init__(self, sampling_options: SamplingOptions):
        self._sampling_options = sampling_options

    @abstractmethod
    def _validate_sampling_options(self):
        return NotImplemented

    @abstractmethod
    def sample(
        self,
        keys_df: DataFrame,
        keys_df_count: int,
        key_columns: list[str],
        target_table: DataFrame,
    ) -> DataFrame:
        return NotImplemented


class RandomSampler(Sampler):
    """
    RandomSampler is a subclass of Sampler that performs random sampling on a given DataFrame.
    """

    def __init__(self, sampling_options: SamplingOptions, seed: int = 100):
        """
        Attributes:
        sampling_options (SamplingOptions): The options for sampling, including method, specifications.
        seed (int): The seed for random sampling to ensure reproducibility.
        """
        super().__init__(sampling_options)
        self.seed = seed

    def _validate_sampling_options(self):
        if self._sampling_options.method != SamplingOptionMethod.RANDOM:
            raise ValueError("RandomSampler: Only 'random' sampling method is supported")

        specs = self._sampling_options.specifications
        if specs.type == SamplingSpecificationsType.COUNT and (
            specs.value is None or (specs.value < _MIN_SAMPLE_COUNT)
        ):
            logger.info(
                f"RandomSampler: Sample count must be >= {_MIN_SAMPLE_COUNT}, " f"flooring to {_MIN_SAMPLE_COUNT}"
            )
            self._sampling_options.specifications.value = _MIN_SAMPLE_COUNT

        elif specs.type == SamplingSpecificationsType.COUNT and specs.value > _MAX_SAMPLE_COUNT:
            logger.info(
                f"RandomSampler: Sample count must be <= {_MAX_SAMPLE_COUNT}, " f"capping to {_MAX_SAMPLE_COUNT}"
            )
            self._sampling_options.specifications.value = _MAX_SAMPLE_COUNT

    def sample(
        self, keys_df: DataFrame, keys_df_count: int, key_columns: list[str], target_table: DataFrame
    ) -> DataFrame:
        """
        Performs random sampling on the given DataFrame based on the specified options.
        - Validates the sampling options.
        - Uses pre-calculated `keys_df_count` from `reconcile_output.mismatch_count` to avoid from recomputing `keys_df`.
        - If the specifications type is FRACTION, samples the DataFrame based on the fraction value.
        - If the specifications type is COUNT, calculates the fraction and samples the DataFrame accordingly,
          then limits the sample size to the specified count.
        - Returns the sampled DataFrame.
        """

        self._validate_sampling_options()
        specs = self._sampling_options.specifications

        default_sampled_df = keys_df.limit(_MIN_SAMPLE_COUNT)

        if specs.type == SamplingSpecificationsType.FRACTION:
            sampled_df = keys_df.sample(fraction=specs.value, seed=self.seed)
        elif specs.type == SamplingSpecificationsType.COUNT:
            total_count = keys_df_count
            sample_size = int(specs.value)
            fraction = min(1.0, sample_size / total_count)
            sampled_df = keys_df.sample(fraction=fraction, seed=self.seed).limit(sample_size)
        else:
            return default_sampled_df

        return sampled_df


class StratifiedSampler(Sampler):
    """
    StratifiedSampler is a subclass of Sampler that performs stratified sampling on a given DataFrame.
    """

    def __init__(self, sampling_options: SamplingOptions, seed: int = 100):
        """
        Attributes:
        sampling_options (SamplingOptions): The options for sampling, including method, specifications,
                                            stratified columns, and stratified buckets.
        seed (int): The seed for random sampling to ensure reproducibility.
        """
        super().__init__(sampling_options)
        self.seed = seed

    def _validate_sampling_options(self):
        if self._sampling_options.method != SamplingOptionMethod.STRATIFIED:
            raise ValueError("StratifiedSampler: Only 'stratified' sampling method is supported")

        specs = self._sampling_options.specifications
        stratified_buckets = self._sampling_options.stratified_buckets

        if specs.type == SamplingSpecificationsType.COUNT and (
            specs.value is None or (specs.value < _MIN_SAMPLE_COUNT)
        ):
            logger.info(
                f"StratifiedSampler: Sample count must be >= {_MIN_SAMPLE_COUNT}, " f"flooring to {_MIN_SAMPLE_COUNT}"
            )
            self._sampling_options.specifications.value = _MIN_SAMPLE_COUNT

        elif specs.type == SamplingSpecificationsType.COUNT and specs.value > _MAX_SAMPLE_COUNT:
            logger.info(
                f"StratifiedSampler: Sample count must be <= {_MAX_SAMPLE_COUNT}, " f"capping to {_MAX_SAMPLE_COUNT}"
            )
            self._sampling_options.specifications.value = _MAX_SAMPLE_COUNT

        if stratified_buckets < _MIN_BUCKET_LIMIT:
            logger.info(
                f"StratifiedSampler: Stratified buckets must be >= {_MIN_BUCKET_LIMIT}, "
                f"flooring to {_MIN_BUCKET_LIMIT}"
            )
            self._sampling_options.stratified_buckets = _MIN_BUCKET_LIMIT
        elif stratified_buckets > _MAX_BUCKET_LIMIT:
            logger.info(
                f"StratifiedSampler: Stratified buckets must be <= {_MAX_BUCKET_LIMIT}, "
                f"capping to {_MAX_BUCKET_LIMIT}"
            )
            self._sampling_options.stratified_buckets = _MAX_BUCKET_LIMIT

    def sample(
        self, keys_df: DataFrame, keys_df_count: int, key_columns: list[str], target_table: DataFrame
    ) -> DataFrame:
        """
        Performs stratified sampling on the given DataFrame based on the specified options.
        - Joins the keys_df with the target_table on the key_columns.
        - Creates a hash bucket column based on the stratified columns.
        - Adds the bucket column to the joined DataFrame.
        - If specifications type is FRACTION, calculates fractions for each bucket and samples accordingly.
        - If specifications type is COUNT, calculates fractions for each bucket and samples accordingly,
          then limits the sample size to the specified count.
        - Returns the sampled DataFrame.
        """
        self._validate_sampling_options()

        specs = self._sampling_options.specifications
        stratified_columns = self._sampling_options.stratified_columns
        non_key_stratified_columns = [
            col for col in (self._sampling_options.stratified_columns or []) if col not in key_columns
        ]
        stratified_buckets = self._sampling_options.stratified_buckets or _MIN_BUCKET_LIMIT

        keys_df.select(*key_columns)
        default_sampled_df = keys_df.limit(_MIN_SAMPLE_COUNT)

        # Join the mismatched_df with target_table_df
        joined_df = keys_df.join(
            target_table, [keys_df[col] == target_table[col] for col in key_columns], "inner"
        ).select(*[keys_df[col] for col in key_columns], *[target_table[col] for col in non_key_stratified_columns])

        # Create a hash bucket column based on the stratified columns
        bucket_col = F.pmod(F.abs(F.hash(*[F.col(c) for c in (stratified_columns or [])])), stratified_buckets)

        # Add the bucket column to the joined_df
        bucketed_df = joined_df.withColumn("bucket", bucket_col)

        if specs.type == SamplingSpecificationsType.FRACTION:
            # Calculate fractions for each bucket
            unique_values = bucketed_df.select("bucket").distinct().collect()
            fractions = {row["bucket"]: specs.value for row in unique_values}
            sampled_df = bucketed_df.sampleBy("bucket", fractions=fractions, seed=self.seed)

        elif specs.type == SamplingSpecificationsType.COUNT:
            # Calculate fractions for each bucket
            sample_size = int(specs.value)
            bucket_counts = bucketed_df.groupBy("bucket").count().collect()
            total_count = sum(row['count'] for row in bucket_counts)
            fractions = {
                row["bucket"]: min(1.0, (sample_size * row['count'] / total_count) / row['count'])
                for row in bucket_counts
            }
            sampled_df = bucketed_df.sampleBy("bucket", fractions=fractions, seed=self.seed).limit(sample_size)

        else:
            return default_sampled_df.select(*key_columns)

        return sampled_df.select(*key_columns)


# TODO: Move away from SamplerFactory to a context-driven approach
class SamplerFactory:
    @staticmethod
    def get_sampler(sampling_options: SamplingOptions, seed: int = 100) -> Sampler:
        # If no sampling options provided, use default
        if sampling_options is None:
            default_sampling_options = SamplingOptions(
                method=SamplingOptionMethod.RANDOM,
                specifications=SamplingSpecifications(type=SamplingSpecificationsType.COUNT, value=_MIN_SAMPLE_COUNT),
                stratified_columns=None,
                stratified_buckets=None,
            )
            logger.info(
                f"SamplerFactory: No sampling options provided, using default options: " f"{default_sampling_options}"
            )
            sampling_options = default_sampling_options

        else:
            logger.info(f"SamplerFactory: Creating sampler using provided options: " f"{sampling_options}")

        # Use a dictionary-based dispatch for better extensibility
        sampler_map = {SamplingOptionMethod.RANDOM: RandomSampler, SamplingOptionMethod.STRATIFIED: StratifiedSampler}

        # Get the sampler class
        sampler_class = sampler_map.get(sampling_options.method)

        if sampler_class is None:
            raise ValueError(f"SamplerFactory : Unsupported sampling method: {sampling_options.method}")

        return sampler_class(sampling_options, seed)



import logging
from importlib.abc import Traversable

from databricks.labs.lsql.backends import SqlBackend

logger = logging.getLogger(__name__)


class TableDeployment:
    def __init__(self, sql_backend: SqlBackend):
        self._sql_backend = sql_backend

    def deploy_table_from_ddl_file(
        self,
        catalog: str,
        schema: str,
        table_name: str,
        ddl_query_filepath: Traversable,
    ):
        """
        Deploys a table to the given catalog and schema
        :param catalog: The table catalog
        :param schema: The table schema
        :param table_name: The table to deploy
        :param ddl_query_filepath: DDL file path
        """
        query = ddl_query_filepath.read_text()
        logger.info(f"Deploying table {table_name} in {catalog}.{schema}")
        logger.info(f"SQL Backend used for deploying table: {type(self._sql_backend).__name__}")
        self._sql_backend.execute(query, catalog=catalog, schema=schema)



import re
import logging
from datetime import datetime

from pyspark.errors import PySparkException
from pyspark.sql import DataFrame, DataFrameReader, SparkSession
from pyspark.sql.functions import col
from sqlglot import Dialect

from databricks.labs.lakebridge.reconcile.connectors.data_source import DataSource
from databricks.labs.lakebridge.reconcile.connectors.jdbc_reader import JDBCReaderMixin
from databricks.labs.lakebridge.reconcile.connectors.models import NormalizedIdentifier
from databricks.labs.lakebridge.reconcile.connectors.secrets import SecretsMixin
from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.recon_config import JdbcReaderOptions, Schema
from databricks.sdk import WorkspaceClient

logger = logging.getLogger(__name__)

_SCHEMA_QUERY = """SELECT
                     COLUMN_NAME,
                     CASE
                        WHEN DATA_TYPE IN ('int', 'bigint')
                            THEN DATA_TYPE
                        WHEN DATA_TYPE IN ('smallint', 'tinyint')
                            THEN 'smallint'
                        WHEN DATA_TYPE IN ('decimal' ,'numeric')
                            THEN 'decimal(' +
                                CAST(NUMERIC_PRECISION AS VARCHAR) + ',' +
                                CAST(NUMERIC_SCALE AS VARCHAR) + ')'
                        WHEN DATA_TYPE IN ('float', 'real')
                                THEN 'double'
                        WHEN CHARACTER_MAXIMUM_LENGTH IS NOT NULL AND DATA_TYPE IN ('varchar','char','text','nchar','nvarchar','ntext')
                                THEN DATA_TYPE
                        WHEN DATA_TYPE IN ('date','time','datetime', 'datetime2','smalldatetime','datetimeoffset')
                                THEN DATA_TYPE
                        WHEN DATA_TYPE IN ('bit')
                                THEN 'boolean'
                        WHEN DATA_TYPE IN ('binary','varbinary')
                                THEN 'binary'
                        ELSE DATA_TYPE
                    END AS 'DATA_TYPE'
                    FROM
                        INFORMATION_SCHEMA.COLUMNS
                    WHERE
                    LOWER(TABLE_NAME) = LOWER('{table}')
                    AND LOWER(TABLE_SCHEMA) = LOWER('{schema}')
                    AND LOWER(TABLE_CATALOG) = LOWER('{catalog}')
              """


class TSQLServerDataSource(DataSource, SecretsMixin, JDBCReaderMixin):
    _DRIVER = "sqlserver"
    _IDENTIFIER_DELIMITER = {"prefix": "[", "suffix": "]"}

    def __init__(
        self,
        engine: Dialect,
        spark: SparkSession,
        ws: WorkspaceClient,
        secret_scope: str,
    ):
        self._engine = engine
        self._spark = spark
        self._ws = ws
        self._secret_scope = secret_scope

    @property
    def get_jdbc_url(self) -> str:
        # Construct the JDBC URL
        return (
            f"jdbc:{self._DRIVER}://{self._get_secret('host')}:{self._get_secret('port')};"
            f"databaseName={self._get_secret('database')};"
            f"user={self._get_secret('user')};"
            f"password={self._get_secret('password')};"
            f"encrypt={self._get_secret('encrypt')};"
            f"trustServerCertificate={self._get_secret('trustServerCertificate')};"
        )

    def read_data(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        query: str,
        options: JdbcReaderOptions | None,
    ) -> DataFrame:
        table_query = query.replace(":tbl", f"{catalog}.{schema}.{table}")
        with_clause_pattern = re.compile(r'WITH\s+.*?\)\s*(?=SELECT)', re.IGNORECASE | re.DOTALL)
        match = with_clause_pattern.search(table_query)
        if match:
            prepare_query_string = match.group(0)
            query = table_query.replace(match.group(0), '')
        else:
            query = table_query
            prepare_query_string = ""
        try:
            if options is None:
                df = self.reader(query, prepare_query_string).load()
            else:
                options = self._get_jdbc_reader_options(options)
                df = self._get_jdbc_reader(table_query, self.get_jdbc_url, self._DRIVER).options(**options).load()
            return df.select([col(column).alias(column.lower()) for column in df.columns])
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "data", table_query)

    def get_schema(
        self,
        catalog: str | None,
        schema: str,
        table: str,
        normalize: bool = True,
    ) -> list[Schema]:
        """
        Fetch the Schema from the INFORMATION_SCHEMA.COLUMNS table in SQL Server.

        If the user's current role does not have the necessary privileges to access the specified
        Information Schema object, RunTimeError will be raised:
        "SQL access control error: Insufficient privileges to operate on schema 'INFORMATION_SCHEMA' "
        """
        schema_query = re.sub(
            r'\s+',
            ' ',
            _SCHEMA_QUERY.format(catalog=catalog, schema=schema, table=table),
        )
        try:
            logger.debug(f"Fetching schema using query: \n`{schema_query}`")
            logger.info(f"Fetching Schema: Started at: {datetime.now()}")
            df = self.reader(schema_query).load()
            schema_metadata = df.select([col(c).alias(c.lower()) for c in df.columns]).collect()
            logger.info(f"Schema fetched successfully. Completed at: {datetime.now()}")
            return [self._map_meta_column(field, normalize) for field in schema_metadata]
        except (RuntimeError, PySparkException) as e:
            return self.log_and_throw_exception(e, "schema", schema_query)

    def reader(self, query: str, prepare_query_str="") -> DataFrameReader:
        return self._get_jdbc_reader(query, self.get_jdbc_url, self._DRIVER, prepare_query_str)

    def normalize_identifier(self, identifier: str) -> NormalizedIdentifier:
        return DialectUtils.normalize_identifier(
            TSQLServerDataSource._normalize_quotes(identifier),
            source_start_delimiter=TSQLServerDataSource._IDENTIFIER_DELIMITER["prefix"],
            source_end_delimiter=TSQLServerDataSource._IDENTIFIER_DELIMITER["suffix"],
        )

    @staticmethod
    def _normalize_quotes(identifier: str):
        if DialectUtils.is_already_delimited(identifier, '"', '"'):
            identifier = identifier[1:-1]
            identifier = identifier.replace('""', '"')
            identifier = (
                TSQLServerDataSource._IDENTIFIER_DELIMITER["prefix"]
                + identifier
                + TSQLServerDataSource._IDENTIFIER_DELIMITER["suffix"]
            )
            return identifier

        return identifier
